/**
 * @file
 * @brief Configuration Header for the Netsweeper Squid Redirector
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */

#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <syslog.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <regex.h>

#include "redir.h"
#include "protocol.h"
#include "config.h"
#include "request.h"
#include "utils.h"
#include "log.h"
#include "simple_iplist.h"

#define SQUID25SERVERPORT 3435
#define SQUID26SERVERPORT 3429
#define SQUID30SERVERPORT 3428

// Regex that specifies the default invalid username characters that will be replaced
// This should be kept functionally identical to the nsd and the webadmin.
//
// We don't need to double-escape the literal backslash (i.e. put four
// backslashes) because backslashes are literal in POSIX bracket expressions.
#define BADCHARSREGEX "[^A-Za-z0-9_@.+'=\\#|-]"

// timeouts in milliseconds
#define CONNTIMEOUT 2000
#define READTIMEOUT 1500
#define WRITETIMEOUT 1500
#define RECONNECTIONDELAY 10000
#define MAXTIMEOUT 8000
#define MAXTIMEOUTNUM 5

#define MAXREQUESTSIZE 8192
#define MAXREPLYSIZE 8192
#define NCONS 1

/** @brief The maximum number of connections we can create to one server */
#define MAXNCONS 20000
#define MAXREQUESTS 100
#define SQUIDREADTIMEOUT 1000
#define SQUIDWRITETIMEOUT 1000
#define LOADBALANCETYPE 1
#define USERNAMEDOMAINMANGLING 0
#define MAXRETRYTIMES 3
#define QUEUEWAITTIMEOUT 50
#define QUEUEFLUSHDELAY 3000

#define DEFLOGFILE "/var/log/redir.log"
#define DEFLOGLEVEL 0

#define CHUNK_SIZE 64

// This is the max length of the error message we will show if the bad chars
// regex is invalid
#define REGEXERRMSG_SIZE 128


void config_print_usage(char** argv);
void config_print_examples(char** argv);

static void config_init_global_conf(GlobalConf *conf);

/**
 * @brief Prepare variables in and allocate memory for the policy server
 *        connection data structures
 *
 * @param host the hostname of the server we are connecting to
 * @param conf the redirector's current configuration
 *
 * @return the initialized connection
 */
static Connection* config_init_connect_info(char* host, GlobalConf* conf)
{
	Connection * ret;
	int port;
	char* ptr;

	// if the argument is empty we have nothing to do
	if (host == NULL || *host == '\0') {
		return NULL;
	}

	// allocate memory
	ret = malloc(sizeof(Connection));
	if (ret == NULL) {
		return NULL;
	}

	// make a hostname copy
	ret->hostname = strdup(host);
	if (ret->hostname == NULL) {
		free(ret);
		return NULL;
	}

	// if the host has a 'port part' get it
	if ((ptr = strchr(ret->hostname, ':')) != NULL) {
		// replace the separator with terminating zero byte
		*ptr++ = '\0';
		// get the port value
		port = atoi(ptr);
	} else {
		// use the global value
		port = conf->nsp_port;
	}

	// fill other fields
	ret->thread_id = 0;
	ret->state = ST_DISCONNECTED;
	ret->socket = -1;
	ret->port = port;
	ret->global_conf = conf;
	ret->start_lock = NULL;

	return ret;
}

/**
 * @brief Set the default values for the redirector's configuration
 *
 * @param conf the configuration data structure whose members we are initializing
 */
static void config_init_global_conf(GlobalConf *conf)
{
	conf->connect_timeout = CONNTIMEOUT;
	conf->read_timeout = READTIMEOUT;
	conf->write_timeout = WRITETIMEOUT;
	conf->reconnect_delay = RECONNECTIONDELAY;
	conf->fail_close_url = NULL;
	conf->fail_close_host = NULL;
	conf->load_bc_type = LOADBALANCETYPE;
	conf->max_retry_times = MAXRETRYTIMES;
	conf->max_request_size = MAXREQUESTSIZE;
	conf->max_reply_size = MAXREPLYSIZE;
	conf->max_timeout = MAXTIMEOUT;
	conf->max_timeouts_num = MAXTIMEOUTNUM;
	conf->ncons = NCONS;
	conf->squid_read_timeout = SQUIDREADTIMEOUT;
	conf->squid_write_timeout = SQUIDREADTIMEOUT;
	conf->max_requests = MAXREQUESTS;
	conf->input_fd = STDIN_FILENO;
	conf->output_fd = STDOUT_FILENO;
	conf->username_domain_mangling = USERNAMEDOMAINMANGLING;
	conf->bad_username_chars_preg = NULL;
	conf->username_lowercase = 1;
	conf->username_remove_domain = NULL;
	conf->nsp_port = SQUID25SERVERPORT;
	conf->input_protocol = SQUID_25;
	conf->output_protocol = SQUID_25;
	conf->nsp_hosts = NULL;
	conf->nsp_num = 0;
	conf->queue_wait_timeout = QUEUEWAITTIMEOUT;
	conf->queue_flush_delay = QUEUEFLUSHDELAY;
	conf->stderr_log = 0;
	conf->server_concurrency = 0;
	conf->log_level = DEFLOGLEVEL;
	conf->ip2domain = NULL;
	conf->unescape = 0;
}

/*
 * @brief Free a global configuration instance
 *
 * @param GlobalConf *conf, a global configuration instance
 */
void config_free_conf(GlobalConf *conf)
{
	int i;
	if(conf->fail_close_url) {
		free(conf->fail_close_url);
	}
	if(conf->fail_close_host) {
		free(conf->fail_close_host);
	}

	if ( conf->bad_username_chars_preg ) {
		// Free internal regex resources
		regfree(conf->bad_username_chars_preg);
		// Free regex_t
		free(conf->bad_username_chars_preg);
	}

	if(conf->username_remove_domain) {
		free(conf->username_remove_domain);
	}

	if(conf->nsp_hosts) {
		for (i = 0; i < conf->nsp_num; i++) {
			free(conf->nsp_hosts[i]);
		}
		free(conf->nsp_hosts);
	}
	free(conf);
}

/**
 * @brief verify that the given hostname is valid
 *
 * Check that this is a valid hostname specification
 *
 * 1) must be non zero length
 * 2) if it contains a port then the portion before must be
 *    a valid hostname and the part after must be a valid
 *    port
 * 3) there's actually nothing else we can check since
 *    basically any character is valid and there's no way
 *    to tell whether it's supposed to be an IP or a hostname
 * 4) no spaces
 **/
static int config_validate_host (const char *host) {
	const char *colon;
	// No spaces allowed
	if ( strchr(host,' ') != NULL ) {
		return 0;
	}
	// Find port spec if it exists
	colon = strchr(host, ':');
	// Validate port since that's easy
	if ( colon != NULL ) {
		char *endptr;
		errno = 0;
		long port = strtol(colon + 1, &endptr, 10);
		if ( errno != 0 || endptr == colon + 1 || *endptr != '\0' ||
		     port <= 0 || port > 65535 ) {
			return 0;
		}
	}
	else {
		colon = host + strlen (host);
	}

	// Make sure hostname is non-zero
	if ( colon - host == 0 ) {
		return 0;
	}

	// I guess we're good
	return 1;
}
/**
 * @brief parse subnet and domain mapping
 *
 * The buf contains data like 1.2.3.4/16:abc.com;5.6.7.8/24:xyz.com.....
 * It is a list of IPsubnet/domainname separated by ";"
 *
 * The original buf is assumed to be commandline optarg, so it will be edit in place
 */
static void config_ip2domain (char *buf, struct simple_iphead **head) {
	char *sep;
	void *domain;
	char *next;
	char *map;

	if (*head == NULL) {
		*head = simple_iplist_init(NULL, NULL);
		if (!head)
			return ;
	}

	next = buf;
	while ((map = strsep(&next, ";")) != NULL) {
		sep = strchr(map, ':');
		if (!sep)
			continue;

		*sep = '\0';
		domain = strdup(sep+1);
		if (!domain)
			return ;

		simple_iplist_add_string(*head, map, domain);
	}
}

/*
 * @brief Parse command line arguments and initialize a Global Confuiguration instance
 *
 * @param int argc, the number of arguments from command line
 * @param char **argv, arguments from command line
 *
 * @return a Global Confuiguration instance on success, print usage and exit on failure
 */
GlobalConf *config_parse_command_line(int argc, char **argv)
{
	GlobalConf *conf;
	int i, option;
	int ret;

	if (argc < 2) {
		config_print_usage(argv);
	}

	conf = xmalloc (sizeof (GlobalConf));
	config_init_global_conf(conf);

	static struct option long_options[] = {
			{"input-protocol", 1, 0, 'i'},
			{"nsp-port", 1, 0, 'P'},
			{"output-protocol", 1, 0, 'o'},
			{"connections-num", 1, 0, 'n'},
			{"username-domain-mangling", 0, 0, 'u'},
			{"username-bad-chars", 1, 0, 'U'},
			{"username-keep-case", 0, 0, 'k'},
			{"username-remove-domain", 1, 0, 'D'},
			//{"load-balance-type", 1, 0, 'b'},   // not supported right now
			{"fail-close-url", 1, 0, 'f'},
			{"fail-close-host", 1, 0, 's'},
			{"connect-timeout", 1, 0, 't'},
			{"read-timeout", 1, 0, 'r'},
			{"write-timeout", 1, 0, 'w'},
			{"reconnect-delay", 1, 0, 'd'},
			{"max-timeout", 1, 0, 'm'},
			{"max-timeouts-num", 1, 0, 'M'},
			{"max-request-size", 1, 0, 'R'},
			{"stderr-log", 0, 0, 'e'},
			{"server-concurrency", 0, 0, 'c'},
			{"log-level", 1, 0, 'L'},
			{"default-domain", 1, 0, 'H'},
			{"unescape-ident", 0, 0, 'G'},
			{"version", 0, 0, 'v'},
			{"help", 0, 0, 'h'},
			{"help-examples", 0, 0, 'x'},
			{0, 0, 0, 0}
	};
	const char *short_options =   "i:o:n:uU:kf:s:t:r:w:d:m:S:M:R:P:ceL:vhxH:GD:";  // -b not supported right now

	// use getopt to do the processing
	while ((option = getopt_long (argc, argv, short_options, long_options, 0)) != -1) {
		int arg_is_num;
		long arg_num = 0;
		// Assume not a number
		arg_is_num = 0;
		// Pre convert optarg to a long int since most of our args are ints
		if ( optarg != NULL ) {
			char *endptr;
			errno = 0;
			arg_num = strtol(optarg, &endptr, 10);

			// Make sure strtol had no errors and
			// it found at least one digit and
			// the entire argument was digits and doesnt have extra stuff on the end
			if ( errno == 0 && endptr != optarg && *endptr == '\0' ) {
				arg_is_num = 1;
			}
		}

#define INTARGCHK(MIN,MAX,MSG)                                      \
	if ( arg_is_num == 0 || arg_num < (MIN) || arg_num > (MAX) ) {  \
		fprintf (stdout, "param value not int or is out of range "  \
		         "(%d,%d): " MSG "\n", MIN, MAX);                   \
		config_print_usage(argv);                                   \
	}

		switch (option) {
			case 'n':
				INTARGCHK(1,MAXNCONS,"number of connections");
				conf->ncons = arg_num;
				break;
			case 'u':
				conf->username_domain_mangling = 1;
				break;
			case 'U':
				conf->bad_username_chars_preg = malloc(sizeof(regex_t));

				// Make sure the preg was properly allocated
				if ( conf->bad_username_chars_preg == NULL ) {
					fprintf(stderr, "Failed to allocate configuration memory\n");
					exit(EXIT_FAILURE);
				}

				// Compile the specified regular expression
				ret = regcomp(conf->bad_username_chars_preg, optarg, REG_EXTENDED);

				// It is invalid
				if ( ret != 0 ) {
					char error[REGEXERRMSG_SIZE];

					// Get the exact regex compilation error
					regerror(ret, conf->bad_username_chars_preg, error, REGEXERRMSG_SIZE);

					printf("Invalid regular expression: %s\n", error);

					// Free any internal regex memory
					regfree(conf->bad_username_chars_preg);

					free(conf->bad_username_chars_preg);

					config_print_usage(argv);
				}

				break;
			case 'k':
				conf->username_lowercase = 0;
				break;
			case 'b':
				INTARGCHK(0,1,"load balance type");
				conf->load_bc_type = arg_num;
				break;
			case 'f':
				// Validate the url a little.
				// It's gotta start with 'http' and it must
				// have a '://' with chars after it
				// And no spaces
				if ( strncmp(optarg, "http", 4) != 0 ||
				     strstr(optarg, "://") == NULL ||
				     strlen(strstr(optarg, "://")) == 3 ) {
					printf ("Invalid fail close url: '%s'\n", optarg);
					config_print_usage(argv);
				}
				conf->fail_close_url = xstrdup(optarg);

				// Set the fail close host to the url if no -s option
				// was specified yet.  We force the port to :443 in this case
				// since we must have something
				if ( conf->fail_close_host == NULL ) {
					char* ptr;
					ptr=strstr(optarg, "://");
					if ( ! ptr ) {
						ptr=optarg;
					}

					conf->fail_close_host = malloc(strlen(ptr+3) + 5);
					strcpy(conf->fail_close_host, ptr+3);
					// truncate the path
					ptr=strchr(conf->fail_close_host, '/');
					if ( ptr ) {
						*ptr='\0';
					}
					// truncate the port
					ptr=strchr(conf->fail_close_host, ':');
					if ( ptr ) {
						*ptr='\0';
					}
					// add the port of :443
					strcat(conf->fail_close_host, ":443");
				}
				break;
			case 's':
				// This will set the fail close host for non-http protocols
				// Format is host:port.  There should be no scheme attached to the
				// remote host.  If -f is used, we use the host in -f with the port
				// of :443, see above
				if ( strstr(optarg, "/") != NULL ) {
					printf ("Invalid fail close host: '%s'\n", optarg);
					config_print_usage(argv);
				}
				// also set by -f above
				if ( conf->fail_close_host ) {
					free(conf->fail_close_host);
				}
				conf->fail_close_host = malloc(strlen(optarg) + 5);
				strcpy(conf->fail_close_host, optarg);
				// add the port of :443 is we have no :
				if ( ! strchr(conf->fail_close_host, ':') ) {
					strcat(conf->fail_close_host, ":443");
				}
				break;

			case 't':
				INTARGCHK(2,20000,"connect timeout");
				conf->connect_timeout = arg_num;
				break;
			case 'r':
				INTARGCHK(2,20000,"read timeout");
				conf->read_timeout = arg_num;
				break;
			case 'w':
				INTARGCHK(2,20000,"write timeout");
				conf->write_timeout = arg_num;
				break;
			case 'd':
				INTARGCHK(0,100000,"reconnect delay");
				conf->reconnect_delay = arg_num;
				break;
			case 'm':
				INTARGCHK(1,100000,"max timeout");
				conf->max_timeout = arg_num;
				break;
			case 'R':
				INTARGCHK(20,100000,"max request size");
				conf->max_request_size = arg_num;
				conf->max_reply_size = arg_num;
				break;
			case 'M':
				INTARGCHK(0,50,"max timeouts num");
				conf->max_timeouts_num = arg_num;
				break;
			case 'e':
				conf->stderr_log = 1;
				break;
			case 'c':
				conf->server_concurrency = 1;
				break;
			case 'L':
				INTARGCHK(0,3,"log level");
				conf->log_level = arg_num;
				break;
			case 'o':
				if ( protocol_config_nametoprotocol(optarg, &conf->output_protocol) != 0 ) {
					config_print_usage(argv);
				}
				break;
			case 'P':
				INTARGCHK(1,65535,"nsp port");
				conf->nsp_port = arg_num;
				break;
			case 'i':
				if ( protocol_config_nametoprotocol(optarg, &conf->input_protocol ) != 0 ) {
					config_print_usage(argv);
				}
				break;
			case 'H':
				config_ip2domain(optarg, &conf->ip2domain) ;
				break;
			case 'G':
				conf->unescape = 1;
				break;
			case 'D':
				if (conf->username_remove_domain != NULL) {
					free(conf->username_remove_domain);
				}
				conf->username_remove_domain = xstrdup(optarg);
				conf->username_remove_domain_len = strlen(conf->username_remove_domain);
				break;
			case 'v':
				printf("Netsweeper Redirector " VERSION " Build: " __DATE__ " " __TIME__ "\n");
				return NULL;
			case 'h':
			case '?':
				config_print_usage(argv);
				break;
			case 'x':
				config_print_examples(argv);
				break;
			default:
				break;
		}
	}
#undef INTARGCHK

	if(argc - optind < 1) {
		config_print_usage(argv);
	}
	conf->nsp_num = argc - optind;
	conf->nsp_hosts = (char **)malloc(sizeof(char *) * (conf->nsp_num));
	for(i = 0; i < conf->nsp_num; i++) {
		char *host = argv[optind + i];
		if ( config_validate_host(host) == 0 ) {
			printf ("Invalid host: '%s'\n", host);
			config_print_usage(argv);
		}
		conf->nsp_hosts[i] = xstrdup(host);
	}

	// The user didn't specify a bad characters regex so we'll use the default
	if ( conf->bad_username_chars_preg == NULL ) {
		conf->bad_username_chars_preg = malloc(sizeof(regex_t));

		// Make sure the preg was properly allocated
		if ( conf->bad_username_chars_preg == NULL ) {
			fprintf(stderr, "Failed to allocate configuration memory\n");
			exit(EXIT_FAILURE);
		}

		// Compile the specified regular expression
		// We can't print errors, so our default MUST have none
		regcomp(conf->bad_username_chars_preg, BADCHARSREGEX, REG_EXTENDED);
	}

	return conf;
}

/*
 * @brief Initialize connections to NSP servers
 *
 * @param GlobalConf *conf, a global configuration instance
 * @int *connectins_num, int reference to the number of connections to valid and resolvable NSP hosts
 *
 * @return a list of initialized connections on success. The number of connection will be passed back to caller.
 * 	Log error message and exit program on failure.
 */
Connection ** config_init_connections(GlobalConf *conf, int *connections_num)
{
	Connection *conn;
	Connection** connection_list = malloc(sizeof(Connection) * CHUNK_SIZE);
	int i, j;
	int num = 0;

	if (connection_list == NULL) {
		log_send(LOG_ERR, "Can not allocate memory\n");
		exit(0);
	}

	if(conf == NULL || conf->nsp_hosts == NULL) {
		log_send(LOG_ERR, "No NSP hosts configured\n");
		exit(0);
	}

	for (i = 0; i < conf->nsp_num; i++) {
		for (j = 0; j < conf->ncons; j++) {
			conn = config_init_connect_info(conf->nsp_hosts[i], conf);
			if (conn == NULL) {
				log_send(LOG_ERR, "Can not init connection to '%s'\n", conf->nsp_hosts[i]);
				break;
			} else {
				// extend the connection list if it's needed
				if (num > 0 && (num % CHUNK_SIZE == 0) ) {
					connection_list = realloc(connection_list, (num + CHUNK_SIZE) * sizeof(Connection));
					if (connection_list == NULL) {
						log_send(LOG_ERR, "Can not allocate memory\n");
						exit(0);
					}
				}
				connection_list[num++] = conn;
			}
		}
	}

	*connections_num = num;;
	return connection_list;
}

void config_free_connections(Connection **connection_list, int num)
{
	int i;

	for (i = 0; i < num; i++) {
		if (connection_list[i] != NULL) {
			if (connection_list[i]->hostname != NULL) {
				free(connection_list[i]->hostname);
			}
			free(connection_list[i]);
		}
	}

	free(connection_list);
}

/*
 * @brief print Netsweeper Redirector usage
 *
 * @param char** argv, command line arguments to run Netsweeper Redirector
 */
void config_print_usage(char** argv)
{
	const char *programname = argv[0];
	GlobalConf *defaults;
	defaults = xmalloc (sizeof (GlobalConf));
	config_init_global_conf(defaults);

	printf("Netsweeper Redirector " VERSION " Build: " __DATE__ " " __TIME__ "\n");
	printf("Usage: %s [options] NSP...\n", programname);
	printf("\t -i, --input-protocol=PROTOCOL        : Specify program input protocol. squid2.4, squid2.5, squid2.6, squid2.7, squid3.0, squid3.1, and squid3.2 are supported (DEFAULT %s)\n", protocol_config_protocoltoname(defaults->input_protocol));
	printf("\t -o, --nsp-protocol=PROTOCOL          : Specify program output protocol to the Netsweeper Policy Servers. squid2.4, squid2.5, squid2.6, squid3.0, squid3.1, and squid3.2 are supported (DEFAULT %s)\n", protocol_config_protocoltoname(defaults->output_protocol));
	printf("\t -P, --nsp-port=PORT                  : Specify the default port the Netsweeper Policy Server listens on. (DEFAULT %d)\n", defaults->nsp_port);
	printf("\t -n, --connections-num=NUM            : Number of concurrent connections per server, default %d, max = %d\n", NCONS, MAXNCONS);
	printf("\t -u, --username-domain-mangling       : Enable username domain mangling, which replaces 'domain\\user' with 'user@domain' (DEFAULT %s)\n", defaults->username_domain_mangling ? "enabled" : "disabled");
	printf("\t -U, --username-bad-chars=REGEX       : Specify the regular expression that defines which characters in a username are invalid (DEFAULT '" BADCHARSREGEX "')\n");
	printf("\t -k, --username-keep-case             : Keep username letters case (DEFAULT: turn username to lower case)\n");
	printf("\t -D, --username-remove-domain=DOMAIN  : Specify the domain that should be removed from username. Empty string ('') removes any domain. Only effective when --username-domain-mangling is on.\n");
	//printf("\t -b, --load-balance-type=B_TYPE       : load balance type 0 = primary/secondary, 1 = roundrobin\n");  // Not supported now
	printf("\t -f, --fail-close-url=URL             : Fail Close to the specified URL (DEFAULT: %s)\n", defaults->fail_close_url == NULL ? "fail open" : defaults->fail_close_url);
	printf("\t -s, --fail-close-host=HOST:PORT      : Fail Close to the host:port for non-http based requests (DEFAULT: %s)\n", defaults->fail_close_host == NULL ? "Fail Open, or HOST of -f argument port of 443" : defaults->fail_close_host);
	printf("\t -t, --connect-timeout=TIMEOUT        : Connect timeout in miliseconds (DEFAULT %d)\n", defaults->connect_timeout);
	printf("\t -r, --read-timeout=TIMEOUT           : Read timeout in miliseconds (DEFAULT %d)\n", defaults->read_timeout);
	printf("\t -w, --write-timeout=TIMEOUT          : Write timeout in miliseconds (DEFAULT %d)\n", defaults->write_timeout);
	printf("\t -d, --reconnect-delay=DELAY          : Delay between attempts to connect when the peer is down (DEFAULT %d)\n", defaults->reconnect_delay);
	printf("\t -m, --max-timeout=TIMEOUT            : max value of timeout to shutdown  connection (DEFAULT %d)\n", defaults->max_timeout);
	printf("\t -M, --max-timeouts-num=NUM           : max number of consecutive timeout errors to shutdown connection (DEFAULT %d)\n", defaults->max_timeouts_num);
	printf("\t -R, --max-request-size=SIZE          : max number of chars in a squid request (DEFAULT %d)\n", defaults->max_request_size);
	printf("\t -e, --stderr-log                     : Print log messages to stderr (DEFAULT %s)\n", defaults->stderr_log ? "log to stderr too" : "use only syslog");
	printf("\t -c, --server-concurrency             : Send squid request id to server (DEFAULT %s)\n", defaults->server_concurrency ? "send it" : "don't send");
	printf("\t -L, --log-level=LEVEL                : The level of messages that should be logged (0 none, 1 errors, 2 warnings, 3 info) (DEFAULT %d)\n", defaults->log_level);
	printf("\t -H, --default-domain                 : IP to domain name mapping for username without the @domain postfix\n");
	printf("\t -G, --unescape                       : URL un-escape the user ident from squid, only effective when --username-domain-mangling is on\n");
	printf("\t -h, --help                           : Display this message\n");
	printf("\t -x, --help-examples                  : Display some usage examples\n");
	printf("\t -v, --version                        : Display version info\n\n");
	printf("\t NSP[:port]  : Netsweeper Policy Server listning on NSP port (DEFAULT: %d)\n\n", defaults->nsp_port);

	printf("Copyright (c) Netsweeper Inc.  All rights reserved.\n\n\n");

	printf("Examples:\n\n");

	printf(" Connect to policy server \"policy.company.com\".  Redirect users to \"http://denypage.company.com\" if \"policy.company.com\" goes down\n");
	printf("   %s -f http://denypage.company.com policy.company.com\n\n", programname);
	printf(" Load balance between two servers. Balance requests evenly, use one exclusively if the other fails\n");
	printf("   %s policy1.company.com policy2.company.com\n", programname);
	config_free_conf(defaults);
	exit(EXIT_FAILURE);
}

/*
 * @brief print Netsweeper Redirector examples of usage
 *
 * @param char** argv, command line arguments to run Netsweeper Redirector
 */
void config_print_examples(char** argv)
{
	const char *programname = argv[0];

	printf (
		"In the examples below it is assumed that your redir (%s) is                        \n"
		"/usr/local/netsweeper/bin/redir.  Update these examples accordingly if your redir  \n"
		"has a different filename.                                                          \n"
		"                                                                                   \n"
		"The Netsweeper Redirector acts as an intermediary between a Squid system and one   \n"
		"(or more) Netsweeper policy server(s).  The redirector is highly configurable in   \n"
		"order to handle a wide range of network configurations, traffic loads, and Squid   \n"
		"versions.                                                                          \n"
		"                                                                                   \n"
		"Netsweeper Redirector Command Line                                                 \n"
		"====================                                                               \n"
		"                                                                                   \n"
		"Basic configuration, running Squid v2.5 with very low load against a basic         \n"
		"Netsweeper policy server that is running squid2.5 request protocol                 \n"
		"(policy1.intranet):                                                                \n"
		"                                                                                   \n"
		"    redir policy1.intranet                                                         \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"Balance load between two policy servers (policy1 & policy2), both of which are     \n"
		"running squid2.5 request protocol (the redir default) on a non-standard port (by   \n"
		"editing nsd.conf):                                                                 \n"
		"                                                                                   \n"
		"    redir -P9000 policy1.intranet policy2.intranet                                 \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"Basic configuration, running Squid v3.2 against a Netsweeper policy server         \n"
		"running squid2.6 request protocol (which uses port 3429 by default in nsd.conf):   \n"
		"                                                                                   \n"
		"    redir -i squid3.2 -o squid2.6 policy1.intranet:3429                            \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"Balance the load between three policy servers, all of which are running squid2.5   \n"
		"request protocol, one of which is using a non-standard port:                       \n"
		"                                                                                   \n"
		"    redir policy1.intranet policy2.intranet policy3.intranet:9000                  \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"The redir is running on a server with at least 4gigs of ram (required for high     \n"
		"values of -n).  Squid request rate is very high, so lots of connections            \n"
		"should be opened to the policy server.  Open 1000 parallel connections to the      \n"
		"policy server:                                                                     \n"
		"                                                                                   \n"
		"    redir -n 1000 policy1.intranet                                                 \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"The policy servers are connected on a fast network and rapid failover is desired   \n"
		"if one policy server goes down.  Specify 200ms read timeout:                       \n"
		"                                                                                   \n"
		"    redir -r 200 policy1.intranet policy2.intranet                                 \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"If all three policy servers go down then users should be allowed to visit the      \n"
		"site they requested (FAIL OPEN):                                                   \n"
		"                                                                                   \n"
		"    redir policy1.intranet policy2.intranet policy3.intranet                       \n"
		"                                                                                   \n"
		"As above but FAIL CLOSE, redirecting users to an explanation/deny page:            \n"
		"                                                                                   \n"
		"    redir -f http://webserver.intranet/deny/redir_outage.html policy1.intranet policy2.intranet policy3.intranet\n"
		"                                                                                   \n"
		"                                                                                   \n"
		"Recommended squid.conf                                                             \n"
		"=====================                                                              \n"
		"                                                                                   \n"
		"NOTE: the Squid configuration on a Netsweeper installation of Squid may include    \n"
		"Redirector parameters at the bottom of the file.  Be sure to make your             \n"
		"customizations there.                                                              \n"
		"                                                                                   \n"
		"                                                                                   \n"
		"# Enter your customized redir command here, including the full path to the         \n"
		"# Netsweeper Redirector. For example:                                              \n"
		"                                                                                   \n"
		"url_rewrite_program /usr/local/netsweeper/bin/redir policy1.intranet               \n"
		"                                                                                   \n"
		"# Starting the redir with multiple policy servers or a value of -n higher than     \n"
		"# one is equivalent to specifying url_rewrite_children > 1.  You will get better   \n"
		"# load balancing and fault tolerance if you specify 1 here and use a high value    \n"
		"# of -n.  INCREASE this value if you are using a Squid version that does not       \n"
		"# support concurrent requests to rewriter.                                         \n"
		"                                                                                   \n"
		"url_rewrite_children 1                                                             \n"
		"                                                                                   \n"
		"# Using Squid concurrency (available in versions > 2.5) allows the Netsweeper      \n"
		"# Redirector to do load balancing and serve multiple requests at once.  If this    \n"
		"# value is 0, you will still get fault tolerance features but will not see load    \n"
		"# balancing benefits or high request rates.                                        \n"
		"# Set this equal to the total number of concurrent policy server connections       \n"
		"# per redir (ie your -n value multiplied by your number of servers)                \n"
		"                                                                                   \n"
		"url_rewrite_concurrency 64                                                         \n"
		"                                                                                   \n"
		"# Turn bypass on if you would prefer to have Squid \"fail open\" if the request    \n"
		"# rate spikes higher than your Netsweeper policy server(s) can handle.  Turn       \n"
		"# bypass off to guarantee that users will not access even in cases of extremely    \n"
		"# high load (Squid will not serve a deny page in this failure case)                \n",
		programname);

	exit(EXIT_FAILURE);
}
