/**
 * @file
 * @brief Implementation of logging module
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */


#define _BSD_SOURCE
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <syslog.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "redir.h"
#include "protocol.h"
#include "config.h"
#include "log.h"

// the log level
static int log_level = 0;
// Since the log_send can be called from concurrent threads we need a mutex
pthread_mutex_t log_lock = PTHREAD_MUTEX_INITIALIZER;

/*
 * @brief Init the logging module
 *
 * @param conf The global config structure
 */
void log_init(GlobalConf* conf)
{
	int options = LOG_PID;

	if (conf->stderr_log) {
		options |= LOG_PERROR;
	}

	openlog ("redir", options, LOG_DAEMON);

	log_level = conf->log_level;
}

/*
 * @brief Write an error message
 *
 * @param type The level of logged message
 * @param message The error message
 */
void log_send(int type, char* message, ...)
{
	va_list args;

	// Don't log the message if its level is not selected for logging
	switch (log_level) {
		case 0:
			// Log no messages
			return;
		case 1:
			// Log only error messages
			if (type != LOG_ERR) {
				return;
			}
			break;
		case 2:
			// Log errors and warnings
			if (type != LOG_ERR && type != LOG_WARNING) {
				return;
			}
			break;
		case 3:
			// Log everything
		default:
			break;
	}

	// since the procedure can be called from concurrent threads take the lock
	pthread_mutex_lock(&log_lock);

	// log the message
	va_start(args, message);
	vsyslog(type, message, args);
	va_end(args);

	// unlock the logfile
	pthread_mutex_unlock(&log_lock);
}
