/*
 * $Id: patricia.c,v 1.7 2005/12/07 20:46:41 dplonka Exp $
 * Dave Plonka <plonka@doit.wisc.edu>
 *
 * This product includes software developed by the University of Michigan,
 * Merit Network, Inc., and their contributors.
 *
 * This file had been called "radix.c" in the MRT sources.
 *
 * I renamed it to "patricia.c" since it's not an implementation of a general
 * radix trie.  Also I pulled in various requirements from "prefix.c" and
 * "demo.c" so that it could be used as a standalone API.
 */

#include <assert.h> /* assert */
#include <ctype.h> /* isdigit */
#include <errno.h> /* errno */
#include <math.h> /* sin */
#include <stddef.h> /* NULL */
#include <stdio.h> /* sprintf, fprintf, stderr */
#include <stdlib.h> /* free, atol, calloc */
#include <string.h> /* memcpy, strchr, strlen */
#include <sys/types.h> /* BSD: for inet_addr */
#include <sys/socket.h> /* BSD, Linux: for inet_addr */
#include <netinet/in.h> /* BSD, Linux: for inet_addr */
#include <arpa/inet.h> /* BSD, Linux, Solaris: for inet_addr */

#include "patricia.h"

/* prefix_tochar
 * convert prefix information to bytes
 */
static u_char *
prefix_tochar (prefix_t * prefix)
{
    if (prefix == NULL)
	return (NULL);

    return ((u_char *) & prefix->add.sin);
}

static int
comp_with_mask (void *addr, void *dest, u_int mask)
{

    if ( /* mask/8 == 0 || */ memcmp (addr, dest, mask / 8) == 0) {
	int n = mask / 8;
	int m = ((-1) << (8 - (mask % 8)));

	if (mask % 8 == 0 || (((u_char *)addr)[n] & m) == (((u_char *)dest)[n] & m))
	    return (1);
    }
    return (0);
}

/*
 * convert prefix information to ascii string with length
 * thread safe and (almost) re-entrant implementation
 */
char *
prefix_toa2x (prefix_t *prefix, char *buff, int with_len)
{
    if (prefix == NULL)
	return ("(Null)");
    assert (prefix->ref_count >= 0);
    if (buff == NULL) {

        struct buffer {
            char buffs[16][48+5];
            u_int i;
        } *buffp;

#    if 0
	THREAD_SPECIFIC_DATA (struct buffer, buffp, 1);
#    else
        { /* for scope only */
	   static struct buffer local_buff;
           buffp = &local_buff;
	}
#    endif
	if (buffp == NULL) {
	    /* XXX should we report an error? */
	    return (NULL);
	}

	buff = buffp->buffs[buffp->i++%16];
    }
    if (prefix->family == AF_INET) {
	u_char *a;
	assert (prefix->bitlen <= 32);
	a = prefix_touchar (prefix);
	if (with_len) {
	    sprintf (buff, "%d.%d.%d.%d/%d", a[0], a[1], a[2], a[3],
		     prefix->bitlen);
	}
	else {
	    sprintf (buff, "%d.%d.%d.%d", a[0], a[1], a[2], a[3]);
	}
	return (buff);
    }
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6) {
	char *r;
	r = (char *) inet_ntop (AF_INET6, &prefix->add.sin6, buff, 48 /* a guess value */ );
	if (r && with_len) {
	    assert (prefix->bitlen <= 128);
	    sprintf (buff + strlen (buff), "/%d", prefix->bitlen);
	}
	return (buff);
    }
#endif /* HAVE_IPV6 */
    else
	return (NULL);
}

/* prefix_toa2
 * convert prefix information to ascii string
 */
char *
prefix_toa2 (prefix_t *prefix, char *buff)
{
    return (prefix_toa2x (prefix, buff, 0));
}

/* prefix_toa
 */
char *
prefix_toa (prefix_t * prefix)
{
    return (prefix_toa2 (prefix, (char *) NULL));
}

prefix_t *
patricia_prefix_new2 (int family, void *dest, int bitlen, prefix_t *prefix)
{
    int dynamic_allocated = 0;
    int default_bitlen = 32;

#ifdef HAVE_IPV6
    if (family == AF_INET6) {
        default_bitlen = 128;
	if (prefix == NULL) {
            prefix = calloc(1, sizeof (prefix6_t));
	    dynamic_allocated++;
	}
	memcpy (&prefix->add.sin6, dest, 16);
    }
    else
#endif /* HAVE_IPV6 */
    if (family == AF_INET) {
		if (prefix == NULL) {
#ifndef NT
            prefix = calloc(1, sizeof (prefix4_t));
#else
			//for some reason, compiler is getting
			//prefix4_t size incorrect on NT
			prefix = calloc(1, sizeof (prefix_t));
#endif /* NT */

			dynamic_allocated++;
		}
		memcpy (&prefix->add.sin, dest, 4);
    }
    else {
        return (NULL);
    }

    prefix->bitlen = (bitlen >= 0)? bitlen: default_bitlen;
    prefix->family = family;
    prefix->ref_count = 0;
    if (dynamic_allocated) {
        prefix->ref_count++;
   }
/* fprintf(stderr, "[C %s, %d]\n", prefix_toa (prefix), prefix->ref_count); */
    return (prefix);
}

prefix_t *
patricia_prefix_new (int family, void *dest, int bitlen)
{
    return (patricia_prefix_new2 (family, dest, bitlen, NULL));
}

prefix_t *
patricia_prefix_ref (prefix_t * prefix)
{
    if (prefix == NULL)
	return (NULL);
    if (prefix->ref_count == 0) {
	/* make a copy in case of a static prefix */
        return (patricia_prefix_new2 (prefix->family, &prefix->add, prefix->bitlen, NULL));
    }
    prefix->ref_count++;
/* fprintf(stderr, "[A %s, %d]\n", prefix_toa (prefix), prefix->ref_count); */
    return (prefix);
}

void
patricia_prefix_deref (prefix_t * prefix)
{
    if (prefix == NULL)
	return;
    /* for secure programming, raise an assert. no static prefix can call this */
    assert (prefix->ref_count > 0);

    prefix->ref_count--;
    assert (prefix->ref_count >= 0);
    if (prefix->ref_count <= 0) {
	free (prefix);
	return;
    }
}

/* } */

/* #define PATRICIA_DEBUG 1 */

static int num_active_patricia = 0;

/* these routines support continuous mask only */

patricia_tree_t *
patricia_init (int maxbits, int (*match_func)(patricia_node_t* node))
{
    patricia_tree_t *patricia = calloc(1, sizeof *patricia);

    patricia->maxbits = maxbits;
    patricia->head = NULL;
    patricia->num_active_node = 0;
    patricia->match_func = match_func;
    assert (maxbits <= PATRICIA_MAXBITS); /* XXX */
    num_active_patricia++;
    return (patricia);
}


/*
 * if func is supplied, it will be called as func(node->data)
 * before deleting the node
 */

void
patricia_free (patricia_tree_t *patricia, void_fn_t func)
{
    assert (patricia);
    if (patricia->head) {

        patricia_node_t *Xstack[PATRICIA_MAXBITS+1];
        patricia_node_t **Xsp = Xstack;
        patricia_node_t *Xrn = patricia->head;

        while (Xrn) {
            patricia_node_t *l = Xrn->l;
            patricia_node_t *r = Xrn->r;

    	    if (Xrn->prefix) {
		patricia_prefix_deref (Xrn->prefix);
		if (Xrn->data && func)
	    	    func (Xrn->data);
    	    }
    	    else {
		assert (Xrn->data == NULL);
    	    }
    	    free (Xrn);
	    patricia->num_active_node--;

            if (l) {
                if (r) {
                    *Xsp++ = r;
                }
                Xrn = l;
            } else if (r) {
                Xrn = r;
            } else if (Xsp != Xstack) {
                Xrn = *(--Xsp);
            } else {
                Xrn = (patricia_node_t *) 0;
            }
        }
    }
    assert (patricia->num_active_node == 0);
    /* free (patricia); */
}


void
patricia_terminate (patricia_tree_t *patricia, void_fn_t func)
{
    patricia_free(patricia, func);
    free (patricia);
    num_active_patricia--;
}


/*
 * if func is supplied, it will be called as func(node->prefix, node->data, arg)
 */

void
patricia_process (patricia_tree_t *patricia, void_fn_t func, void* arg)
{
    patricia_node_t *node;
    assert (func);

    PATRICIA_WALK (patricia->head, node) {
	func (node->prefix, node->data, arg);
    } PATRICIA_WALK_END;
}

size_t patricia_walk(patricia_tree_t *patricia, void_fn_t func, void* arg)
{
	return patricia_walk_inorder(patricia->head, func, arg);
}

size_t
patricia_walk_inorder(patricia_node_t *node, void_fn_t func, void* arg)
{
    size_t n = 0;
    assert(func);

    if (node->l) {
         n += patricia_walk_inorder(node->l, func, arg);
    }

    if (node->prefix) {
	func(node->prefix, node->data, arg);
	n++;
    }

    if (node->r) {
         n += patricia_walk_inorder(node->r, func, arg);
    }

    return n;
}


patricia_node_t *
patricia_search_exact (patricia_tree_t *patricia, prefix_t *prefix)
{
    patricia_node_t *node;
    u_char *addr;
    u_int bitlen;

    assert (patricia);
    assert (prefix);
    assert (prefix->bitlen <= patricia->maxbits);

    if (patricia->head == NULL)
	return (NULL);

    node = patricia->head;
    addr = prefix_touchar (prefix);
    bitlen = prefix->bitlen;

    while (node->bit < bitlen) {

	if (BIT_TEST (addr[node->bit >> 3], 0x80 >> (node->bit & 0x07))) {
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_search_exact: take right %s/%d\n",
	                 prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_search_exact: take right at %d\n",
			 node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->r;
	}
	else {
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_search_exact: take left %s/%d\n",
	                 prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_search_exact: take left at %d\n",
			 node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->l;
	}

	if (node == NULL)
	    return (NULL);
    }

#ifdef PATRICIA_DEBUG
    if (node->prefix)
        fprintf (stderr, "patricia_search_exact: stop at %s/%d\n",
	         prefix_toa (node->prefix), node->prefix->bitlen);
    else
        fprintf (stderr, "patricia_search_exact: stop at %d\n", node->bit);
#endif /* PATRICIA_DEBUG */
    if (node->bit > bitlen || node->prefix == NULL)
	return (NULL);
    assert (node->bit == bitlen);
    assert (node->bit == node->prefix->bitlen);
    if (comp_with_mask (prefix_tochar (node->prefix), prefix_tochar (prefix),
			bitlen)) {
#ifdef PATRICIA_DEBUG
        fprintf (stderr, "patricia_search_exact: found %s/%d\n",
	         prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	return (node);
    }
    return (NULL);
}


/* if inclusive != 0, "best" may be the given prefix itself */
patricia_node_t *
patricia_search_best2 (patricia_tree_t *patricia, prefix_t *prefix, int inclusive)
{
    patricia_node_t *node;
    patricia_node_t *stack[PATRICIA_MAXBITS + 1];
    u_char *addr;
    u_int bitlen;
    int cnt = 0;

    assert (patricia);
    assert (prefix);
    assert (prefix->bitlen <= patricia->maxbits);

    if (patricia->head == NULL)
	return (NULL);

    node = patricia->head;
    addr = prefix_touchar (prefix);
    bitlen = prefix->bitlen;

    while (node->bit < bitlen) {

	if (node->prefix) {
#ifdef PATRICIA_DEBUG
            fprintf (stderr, "patricia_search_best: push %s/%d\n",
	             prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	    stack[cnt++] = node;
	}

	if (BIT_TEST (addr[node->bit >> 3], 0x80 >> (node->bit & 0x07))) {
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_search_best: take right %s/%d\n",
	                 prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_search_best: take right at %d\n",
			 node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->r;
	}
	else {
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_search_best: take left %s/%d\n",
	                 prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_search_best: take left at %d\n",
			 node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->l;
	}

	if (node == NULL)
	    break;
    }

    if (inclusive && node && node->prefix)
	stack[cnt++] = node;

#ifdef PATRICIA_DEBUG
    if (node == NULL)
        fprintf (stderr, "patricia_search_best: stop at null\n");
    else if (node->prefix)
        fprintf (stderr, "patricia_search_best: stop at %s/%d\n",
	         prefix_toa (node->prefix), node->prefix->bitlen);
    else
        fprintf (stderr, "patricia_search_best: stop at %d\n", node->bit);
#endif /* PATRICIA_DEBUG */

    if (cnt <= 0)
	return (NULL);

    while (--cnt >= 0) {
	node = stack[cnt];
#ifdef PATRICIA_DEBUG
        fprintf (stderr, "patricia_search_best: pop %s/%d\n",
	         prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	if (comp_with_mask (prefix_tochar (node->prefix),
			    prefix_tochar (prefix),
			    node->prefix->bitlen)) {
#ifdef PATRICIA_DEBUG
            fprintf (stderr, "patricia_search_best: found %s/%d\n",
	             prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
		// run the custom function on data
		// can expire the entry... how nice...
		if ( patricia->match_func != NULL ) {
			if ( patricia->match_func(node) ) {
				return (node);
			}
		}
		else {
	    		return (node);
		}
	}
    }
    return (NULL);
}


patricia_node_t *
patricia_search_best (patricia_tree_t *patricia, prefix_t *prefix)
{
    return (patricia_search_best2 (patricia, prefix, 1));
}


patricia_node_t *
patricia_lookup (patricia_tree_t *patricia, prefix_t *prefix)
{
    patricia_node_t *node, *new_node, *parent, *glue;
    u_char *addr, *test_addr;
    u_int bitlen, check_bit, differ_bit, i;
    int j, r;

    assert (patricia);
    assert (prefix);
    assert (prefix->bitlen <= patricia->maxbits);

    if (patricia->head == NULL) {
	node = calloc(1, sizeof *node);
	node->bit = prefix->bitlen;
	node->prefix = patricia_prefix_ref (prefix);
	node->parent = NULL;
	node->l = node->r = NULL;
	node->data = NULL;
	patricia->head = node;
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_lookup: new_node #0 %s/%d (head)\n",
		 prefix_toa (prefix), prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	patricia->num_active_node++;
	return (node);
    }

    addr = prefix_touchar (prefix);
    bitlen = prefix->bitlen;
    node = patricia->head;

    while (node->bit < bitlen || node->prefix == NULL) {

	if (node->bit < patricia->maxbits &&
	    BIT_TEST (addr[node->bit >> 3], 0x80 >> (node->bit & 0x07))) {
	    if (node->r == NULL)
		break;
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_lookup: take right %s/%d\n",
	                 prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_lookup: take right at %d\n", node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->r;
	}
	else {
	    if (node->l == NULL)
		break;
#ifdef PATRICIA_DEBUG
	    if (node->prefix)
    	        fprintf (stderr, "patricia_lookup: take left %s/%d\n",
	             prefix_toa (node->prefix), node->prefix->bitlen);
	    else
    	        fprintf (stderr, "patricia_lookup: take left at %d\n", node->bit);
#endif /* PATRICIA_DEBUG */
	    node = node->l;
	}

	assert (node);
    }

    assert (node->prefix);
#ifdef PATRICIA_DEBUG
    fprintf (stderr, "patricia_lookup: stop at %s/%d\n",
	     prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */

    test_addr = prefix_touchar (node->prefix);
    /* find the first bit different */
    check_bit = (node->bit < bitlen)? node->bit: bitlen;
    differ_bit = 0;
    for (i = 0; i*8 < check_bit; i++) {
	if ((r = (addr[i] ^ test_addr[i])) == 0) {
	    differ_bit = (i + 1) * 8;
	    continue;
	}
	/* I know the better way, but for now */
	for (j = 0; j < 8; j++) {
	    if (BIT_TEST (r, (0x80 >> j)))
		break;
	}
	/* must be found */
	assert (j < 8);
	differ_bit = i * 8 + j;
	break;
    }
    if (differ_bit > check_bit)
	differ_bit = check_bit;
#ifdef PATRICIA_DEBUG
    fprintf (stderr, "patricia_lookup: differ_bit %d\n", differ_bit);
#endif /* PATRICIA_DEBUG */

    parent = node->parent;
    while (parent && parent->bit >= differ_bit) {
	node = parent;
	parent = node->parent;
#ifdef PATRICIA_DEBUG
	if (node->prefix)
            fprintf (stderr, "patricia_lookup: up to %s/%d\n",
	             prefix_toa (node->prefix), node->prefix->bitlen);
	else
            fprintf (stderr, "patricia_lookup: up to %d\n", node->bit);
#endif /* PATRICIA_DEBUG */
    }

    if (differ_bit == bitlen && node->bit == bitlen) {
	if (node->prefix) {
#ifdef PATRICIA_DEBUG
    	    fprintf (stderr, "patricia_lookup: found %s/%d\n",
		     prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	    return (node);
	}
	node->prefix = patricia_prefix_ref (prefix);
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_lookup: new node #1 %s/%d (glue mod)\n",
		 prefix_toa (prefix), prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	assert (node->data == NULL);
	return (node);
    }

    new_node = calloc(1, sizeof *new_node);
    new_node->bit = prefix->bitlen;
    new_node->prefix = patricia_prefix_ref (prefix);
    new_node->parent = NULL;
    new_node->l = new_node->r = NULL;
    new_node->data = NULL;
    patricia->num_active_node++;

    if (node->bit == differ_bit) {
	new_node->parent = node;
	if (node->bit < patricia->maxbits &&
	    BIT_TEST (addr[node->bit >> 3], 0x80 >> (node->bit & 0x07))) {
	    assert (node->r == NULL);
	    node->r = new_node;
	}
	else {
	    assert (node->l == NULL);
	    node->l = new_node;
	}
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_lookup: new_node #2 %s/%d (child)\n",
		 prefix_toa (prefix), prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	return (new_node);
    }

    if (bitlen == differ_bit) {
	if (bitlen < patricia->maxbits &&
	    BIT_TEST (test_addr[bitlen >> 3], 0x80 >> (bitlen & 0x07))) {
	    new_node->r = node;
	}
	else {
	    new_node->l = node;
	}
	new_node->parent = node->parent;
	if (node->parent == NULL) {
	    assert (patricia->head == node);
	    patricia->head = new_node;
	}
	else if (node->parent->r == node) {
	    node->parent->r = new_node;
	}
	else {
	    node->parent->l = new_node;
	}
	node->parent = new_node;
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_lookup: new_node #3 %s/%d (parent)\n",
		 prefix_toa (prefix), prefix->bitlen);
#endif /* PATRICIA_DEBUG */
    }
    else {
        glue = calloc(1, sizeof *glue);
        glue->bit = differ_bit;
        glue->prefix = NULL;
        glue->parent = node->parent;
        glue->data = NULL;
        patricia->num_active_node++;
	if (differ_bit < patricia->maxbits &&
	    BIT_TEST (addr[differ_bit >> 3], 0x80 >> (differ_bit & 0x07))) {
	    glue->r = new_node;
	    glue->l = node;
	}
	else {
	    glue->r = node;
	    glue->l = new_node;
	}
	new_node->parent = glue;

	if (node->parent == NULL) {
	    assert (patricia->head == node);
	    patricia->head = glue;
	}
	else if (node->parent->r == node) {
	    node->parent->r = glue;
	}
	else {
	    node->parent->l = glue;
	}
	node->parent = glue;
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_lookup: new_node #4 %s/%d (glue+node)\n",
		 prefix_toa (prefix), prefix->bitlen);
#endif /* PATRICIA_DEBUG */
    }
    return (new_node);
}


void
patricia_remove (patricia_tree_t *patricia, patricia_node_t *node)
{
    patricia_node_t *parent, *child;

    assert (patricia);
    assert (node);

    if (node->r && node->l) {
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_remove: #0 %s/%d (r & l)\n",
		 prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */

	/* this might be a placeholder node -- have to check and make sure
	 * there is a prefix aossciated with it ! */
	if (node->prefix != NULL)
	  patricia_prefix_deref (node->prefix);
	node->prefix = NULL;
	/* Also I needed to clear data pointer -- masaki */
	node->data = NULL;
	return;
    }

    if (node->r == NULL && node->l == NULL) {
#ifdef PATRICIA_DEBUG
	fprintf (stderr, "patricia_remove: #1 %s/%d (!r & !l)\n",
		 prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
	parent = node->parent;
	patricia_prefix_deref (node->prefix);
	free (node);
        patricia->num_active_node--;

	if (parent == NULL) {
	    assert (patricia->head == node);
	    patricia->head = NULL;
	    return;
	}

	if (parent->r == node) {
	    parent->r = NULL;
	    child = parent->l;
	}
	else {
	    assert (parent->l == node);
	    parent->l = NULL;
	    child = parent->r;
	}

	if (parent->prefix)
	    return;

	/* we need to remove parent too */

	if (parent->parent == NULL) {
	    assert (patricia->head == parent);
	    patricia->head = child;
	}
	else if (parent->parent->r == parent) {
	    parent->parent->r = child;
	}
	else {
	    assert (parent->parent->l == parent);
	    parent->parent->l = child;
	}
	child->parent = parent->parent;
	free (parent);
        patricia->num_active_node--;
	return;
    }

#ifdef PATRICIA_DEBUG
    fprintf (stderr, "patricia_remove: #2 %s/%d (r ^ l)\n",
	     prefix_toa (node->prefix), node->prefix->bitlen);
#endif /* PATRICIA_DEBUG */
    if (node->r) {
	child = node->r;
    }
    else {
	assert (node->l);
	child = node->l;
    }
    parent = node->parent;
    child->parent = parent;

    patricia_prefix_deref (node->prefix);
    free (node);
    patricia->num_active_node--;

    if (parent == NULL) {
	assert (patricia->head == node);
	patricia->head = child;
	return;
    }

    if (parent->r == node) {
	parent->r = child;
    }
    else {
        assert (parent->l == node);
	parent->l = child;
    }
}

void* patricia_find_exact_data(patricia_tree_t * tree, in_addr_t ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;
	void * data;


	patricia_prefix_new2(AF_INET, &ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) == NULL) {
		return NULL;
	}
	data = node->data;
	return data;
}

int patricia_find_exact(patricia_tree_t * tree, in_addr_t ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;


	patricia_prefix_new2(AF_INET, &ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) == NULL) {
		return 1;
	}
	return 0;
}

void* patricia_find_data(patricia_tree_t * tree, in_addr_t ip)
{
	prefix_t prefix;
	patricia_node_t *node;
	int cidr = 32;

	patricia_prefix_new2(AF_INET, &ip, cidr, &prefix);
	if ((node = patricia_search_best(tree, &prefix)) == NULL) {
		return NULL;
	}
	return node->data;
}

int patricia_find(patricia_tree_t * tree, in_addr_t ip)
{
	prefix_t prefix;
	patricia_node_t *node;
	int cidr = 32;

	patricia_prefix_new2(AF_INET, &ip, cidr, &prefix);
	if ((node = patricia_search_best(tree, &prefix)) == NULL) {
		return 1;
	}
	return 0;
}

void* patricia_delete(patricia_tree_t* tree, in_addr_t ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;
	void* data;


	patricia_prefix_new2(AF_INET, &ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) != NULL) {
		data=node->data;
		patricia_remove(tree, node);
		return data;
	}
	return NULL;
}

int patricia_add(patricia_tree_t* tree, in_addr_t ip, int cidr, void* data)
{
	prefix_t* prefix;
	patricia_node_t *node;

	prefix = patricia_prefix_new(AF_INET, &ip, cidr);
	node = patricia_lookup(tree, prefix);
	if ( node ) {
		node->data = data;
		patricia_prefix_deref (prefix);
		return 0;
	}
	patricia_prefix_deref (prefix);
	return 1;
}

// IPv6 functions

void* patricia_ipv6_find_exact_data(patricia_tree_t * tree, struct in6_addr *ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;
	void * data;


	patricia_prefix_new2(AF_INET6, ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) == NULL) {
		return NULL;
	}
	data = node->data;
	return data;
}

int patricia_ipv6_find_exact(patricia_tree_t * tree, struct in6_addr *ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;


	patricia_prefix_new2(AF_INET6, ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) == NULL) {
		return 1;
	}
	return 0;
}

void* patricia_ipv6_find_data(patricia_tree_t * tree, struct in6_addr *ip)
{
	prefix_t prefix;
	patricia_node_t *node;
	int cidr = 128;

	patricia_prefix_new2(AF_INET6, ip, cidr, &prefix);
	if ((node = patricia_search_best(tree, &prefix)) == NULL) {
		return NULL;
	}
	return node->data;
}

int patricia_ipv6_find(patricia_tree_t * tree, struct in6_addr *ip)
{
	prefix_t prefix;
	patricia_node_t *node;
	int cidr = 128;

	patricia_prefix_new2(AF_INET6, ip, cidr, &prefix);
	if ((node = patricia_search_best(tree, &prefix)) == NULL) {
		return 1;
	}
	return 0;
}

void* patricia_ipv6_delete(patricia_tree_t* tree, struct in6_addr *ip, int cidr)
{
	prefix_t prefix;
	patricia_node_t *node;
	void* data;

	patricia_prefix_new2(AF_INET6, ip, cidr, &prefix);
	if ((node = patricia_search_exact (tree, &prefix)) != NULL) {
		data=node->data;
		patricia_remove(tree, node);
		return data;
	}
	return NULL;
}

int patricia_ipv6_add(patricia_tree_t* tree, struct in6_addr *ip, int cidr, void* data)
{
	prefix_t* prefix;
	patricia_node_t *node;

	prefix = patricia_prefix_new(AF_INET6, ip, cidr);
	node = patricia_lookup(tree, prefix);
	if ( node ) {
		node->data = data;
		patricia_prefix_deref (prefix);
		return 0;
	}
	patricia_prefix_deref (prefix);
	return 1;
}

#ifdef PATRICIA_MAIN

void patricia_test_lookup(patricia_tree_t* tree, in_addr_t ip, char* string)
{
	char* data=NULL;

	if ( patricia_find(tree, ip) == 0 ) {
		data = patricia_find_data(tree, ip);
		if ( data ) {
			if ( !strcmp(data, string) ) {
				printf("PASSED: %s\n", string);
				return;
			}
		}
	}
	printf("FAILED: to lookup %d with %s got %s\n", ip, string, data);
}
int main()
{
	patricia_tree_t *tree;

	tree = patricia_init(32, NULL);

	// Test the helper functions
	patricia_add(tree, inet_addr("192.168.123.0"), 24, "test1");
	patricia_add(tree, inet_addr("192.168.123.1"), 32, "test1.32");
	patricia_test_lookup(tree, inet_addr("192.168.123.1"), "test1.32");
	patricia_test_lookup(tree, inet_addr("192.168.123.2"), "test1");

	patricia_add(tree, inet_addr("192.168.0.0"), 16, "test2");
	patricia_test_lookup(tree, inet_addr("192.168.123.1"), "test1.32");
	patricia_test_lookup(tree, inet_addr("192.168.123.2"), "test1");
	patricia_test_lookup(tree, inet_addr("192.168.124.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.0.0"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.255"), "test2");


	patricia_add(tree, inet_addr("192.0.0.0"), 8, "test3");
	patricia_test_lookup(tree, inet_addr("192.168.123.1"), "test1.32");
	patricia_test_lookup(tree, inet_addr("192.168.123.2"), "test1");
	patricia_test_lookup(tree, inet_addr("192.168.124.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.0.0"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.255"), "test2");
	patricia_test_lookup(tree, inet_addr("192.0.255.255"), "test3");
	patricia_test_lookup(tree, inet_addr("192.0.0.0"), "test3");
	patricia_test_lookup(tree, inet_addr("192.167.255.255"), "test3");

	patricia_add(tree, 0, 0, "test4");
	patricia_test_lookup(tree, inet_addr("192.168.123.1"), "test1.32");
	patricia_test_lookup(tree, inet_addr("192.168.123.2"), "test1");
	patricia_test_lookup(tree, inet_addr("192.168.124.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.2"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.0.0"), "test2");
	patricia_test_lookup(tree, inet_addr("192.168.255.255"), "test2");
	patricia_test_lookup(tree, inet_addr("192.0.255.255"), "test3");
	patricia_test_lookup(tree, inet_addr("192.0.0.0"), "test3");
	patricia_test_lookup(tree, inet_addr("192.167.255.255"), "test3");
	patricia_test_lookup(tree, inet_addr("123.123.123.1"), "test4");
	patricia_test_lookup(tree, inet_addr("1.1.1.1"), "test4");
	patricia_test_lookup(tree, inet_addr("192.168.123.123"), "test1");
	patricia_test_lookup(tree, inet_addr("192.1.1.1"), "test3");
	patricia_test_lookup(tree, 0, "test4");
	patricia_test_lookup(tree, inet_addr("1.1.1.1"), "test4");


	return 0;
}
#endif
