/**
 * @file
 * @brief Implementation of protocol parsing framework
 *
 * This module contains the framework for selecting between multiple protocols,
 * calling protocol parsing functions and modifying usernames.
 *
 * Functions that parse individual protocols are defined elsewhere and their
 * poinnters are stored in the defined_protocols array, along with their name
 * and internal ID.
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content
 * Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark
 * of Netsweeper Inc.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <regex.h>

#include "simple_iplist.h"
#include "redir.h"
#include "redir.h"
#include "protocol.h"
#include "protocol_def.h"

/********************************************************
 *****************  Internal Data Types *****************
 *******************************************************/

typedef struct {
	protocol_select which_protocol;
	protocol_parser_func *parser;
	protocol_encoder_func *encoder;
	const char *name;
} protocol_internal_definition;


/********************************************************
 **********  Helper Function Declarations ***************
 *******************************************************/

// Allocate and initialize the structure
protocol_input_line *protocol_init (size_t input_str_len);

// find specified protocol in our list of defined protocols
const protocol_internal_definition *protocol_find_definition (protocol_select prot);

/********************************************************
 *****************  Internal Constants ******************
 *******************************************************/

/* This is the count of protocols in the array below. */
static const int defined_protocols_count = 7;

/* This array defines all the protocols we can handle */
static const protocol_internal_definition defined_protocols[] = {
	{.which_protocol = SQUID_24,
	 .parser = &squid24_parse,
	 .encoder = &squid24_encode,
	 .name = "squid2.4"},
	{.which_protocol = SQUID_25,
	 .parser = &squid25_parse,
	 .encoder = &squid25_encode,
	 .name = "squid2.5"},
	{.which_protocol = SQUID_26,
	 .parser = &squid26_parse,
	 .encoder = &squid26_encode,
	 .name = "squid2.6"},
	{.which_protocol = SQUID_27,
	 .parser = &squid27_parse,
	 .encoder = &squid27_encode,
	 .name = "squid2.7"},
	{.which_protocol = SQUID_30,
	 .parser = &squid30_parse,
	 .encoder = &squid30_encode,
	 .name = "squid3.0"},
	{.which_protocol = SQUID_31,
	 .parser = &squid31_parse,
	 .encoder = &squid31_encode,
	 .name = "squid3.1"},
	{.which_protocol = SQUID_32,
	 .parser = &squid32_parse,
	 .encoder = &squid32_encode,
	 .name = "squid3.2"}
};
	

/********************************************************
 **********  Public Function Definitions ***************
 *******************************************************/

/** @brief Convert from a protocol name to its protocol_select number.
 *
 * When the user selects a protocol, they do so by specifying its name. For
 * example, "-o squid2.5" for outputting Squid version 2.5. This function
 * converts from the name to the number we use to identify the protocol
 * internally.
 *
 * @param[in] name the name of the protocol
 * @param[out] protocol the destination where we are going to store the protocol
 *
 * @return 0 on successful conversion from name to protocol
 *        -1 on error
 *
 * If an error occurs, errno will be set to one of the following:
 *        EINVAL - name is invalid (NULL)
 *        ENOENT - no protocol with that name exists
 */
int protocol_config_nametoprotocol(const char *name, protocol_select *protocol)
{
	int i;

	if ( name == NULL ) {
		errno = EINVAL;
		return -1;
	}

	// Iterate through our defined protocols, looking for this name
	for (i = 0; i < defined_protocols_count; i++) {
		if ( strcmp(name, defined_protocols[i].name) == 0 ) {
			// We found the protocol so let's set it
			*protocol = defined_protocols[i].which_protocol;
			return 0;
		}
	}

	// We couldn't find a protocol with that name
	errno = ENOENT;
	return -1;

}


/**
 * @brief Translate a protocol number to its human-readable name
 *
 * @param protocol the number of the protocol we're translating
 *
 * @return the name of the protocol
 *         or "INVALID" if a protocol with that number does not exist
 */
const char *protocol_config_protocoltoname (protocol_select protocol)
{
	const protocol_internal_definition *prot_def;
	prot_def = protocol_find_definition (protocol);
	if (prot_def == NULL || prot_def->name == NULL)
	{
		return "INVALID";
	}

	return prot_def->name;
}


/** @brief Parse an input line of text
 *
 * This function parses the input line of text and
 * ensures that it is in the proper format.  Detects
 * protocol errors.
 *
 * @param inputstr The full input str, we make a copy
 *                 of this string
 * @param parsed_line Parse the input into this structure
 * @param protocol The protocol that we should parse for.
 *                 It will be an error if input string
 *                 doesnt match this protocol
 *
 * @returns 0 on successful parse
 *         -1 on parse or other error
 *            EINVAL - invalid inputsring, protocol, etc
 *            EIO - Parsing error
 *            ENXIO - parser for specified protocol doesnt exist
 *            ENOMEM - Memory allocation error
 **/
int protocol_parse (const char *inputstr, protocol_input_line **parsed_line, protocol_select protocol)
{
	size_t input_len;
	const protocol_internal_definition *prot_def;

	// Check for input errors
	if (inputstr == NULL || parsed_line == NULL)
	{
		errno = EINVAL;
		return -1;
	}	

	// Make sure this protocol has a parser
	prot_def = protocol_find_definition (protocol);
	if (prot_def == NULL || prot_def->parser == NULL)
	{
		errno = ENXIO;
		return -1;
	}
	
	// Create and initialize parsed line structure
	input_len = strlen (inputstr);
	*parsed_line = protocol_init (input_len);
	if (*parsed_line == NULL)
	{
		errno = ENOMEM;
		return -1;
	}
	strcpy ((*parsed_line)->line_buffer, inputstr);
	// Get rid of '\n' if it exists
	if ((*parsed_line)->line_buffer[input_len - 1] == '\n')
	{
		(*parsed_line)->line_buffer[input_len - 1] = '\0';
	}

	// Call the parser for the specific protocol
	if ((*prot_def->parser)(*parsed_line) == 0)
	{
		return 0;
	}
	else
	{
		// The parser had an error, return error
		return -1;
	}
}

/** @brief This writes a line of text in the desired protocol
 *
 * This function takes an parsedline structure containing
 * request information and writes a new line of text containing
 * all the information requried by the protocol in the required format.
 *
 * @param conf Configuraton structure
 * @param parsed_line Read the parsed line from this structure
 * @param line_buffer Write the string to this text buffer
 * @param line_buffer_len Write at most this many characters
 * @param protocol The protocol that we should encode in.
 *
 * @returns 0 on successful encode
 *         -1 on error
 *            EINVAL - bad parameters
 *            ENOENT - a field required by this protocol was not present
 *                     in the parsed line structure
 *            ENXIO - encoder for specified protocol doesnt exist
 *            ENOMEM - the line_buffer is too small for this encoded line
 **/
int protocol_encode (GlobalConf *conf, const protocol_input_line *parsed_line, char *line_buffer, size_t line_buffer_len, protocol_select protocol)
{
	const protocol_internal_definition *prot_def;

	// Check for input errors
	if (parsed_line == NULL || line_buffer == NULL)
	{
		errno = EINVAL;
		return -1;
	}

	// Find the encoder for this protocol
	prot_def = protocol_find_definition (protocol);
	if (prot_def == NULL || prot_def->encoder == NULL)
	{
		errno = ENXIO;
		return -1;
	}
	
	// Call the encoder
	return (*prot_def->encoder)(conf, parsed_line, line_buffer, line_buffer_len);
}


/** @brief Use this to cleanup a parsed line structure
 *
 * You will have a parsed line structure if the _parse function
 * returns SUCCESS.  This structure must be cleaned up by caller
 * by calling this function
 *
 * @param parsed_line The structure to free and clean up
 **/
void protocol_parsed_line_destroy (protocol_input_line *parsed_line)
{
	int i;
	if(parsed_line == NULL) {
		return;
	}

	if(parsed_line->req_id) {
		free (parsed_line->req_id);
	}
	if(parsed_line->url) {
		free (parsed_line->url);
	}
	if(parsed_line->IP) {
		free (parsed_line->IP);
	}
	if(parsed_line->username) {
		free (parsed_line->username);
	}
	if(parsed_line->method) {
		free (parsed_line->method);
	}
	if(parsed_line->group) {
		free (parsed_line->group);
	}
	for (i = 0; i < MAX_PROTOCOLS_OTHER_OPTIONS; i++)
	{
		if(parsed_line->otheroptions[i]) {
			free (parsed_line->otheroptions[i]);
		}
	}

	free (parsed_line);
}

/*
 * Brief: get a two charachor size hex number.
 * -1 if does not look like hex.
 */
static int get_hex(unsigned char ch) {
	if (ch <= '9' && ch >= '0')
		return (ch - '0');
	if (ch <= 'z' && ch >= 'a')
		return (ch - 'a' + 10);
	if (ch <= 'Z' && ch >= 'A')
		return (ch - 'A' + 10);
	return -1;
}

/**
 * Brief: rfc 1738 unescape the string
 * result string is always shorter, so no need to expand size
 **/
static void
protocol_unescape(char *s)
{
    int i, j;
    for (i = j = 0; s[j]; i++, j++) {
	    s[i] = s[j];
	    if (s[j] != '%') {
		    /* do nothing */
	    }
	    else if (s[j + 1] == '%') {   /* %% case */
		    j++;
	    } else {
		    int v1, v2, x;
		    v1 = get_hex(s[j + 1]);
		    if (v1 < 0)
			    continue;
		    v2 = get_hex(s[j + 2]);
		    if (v2 < 0)
			    continue;
		    x = v1 << 4 | v2;
		    if (x > 0 && x <= 255) {
			    s[i] = x;
			    j += 2;
		    }
	    }
    }
    s[i] = '\0';
}


/** @brief Mangle username
 *
 * This implements an option in the redir where we rewrite
 * usernames containing a '\' character and put a @ in there
 * instead (and swap the domain and user)
 *
 * @param parsed_line The parsed line with a username you want to mangle
 *
 * @returns 0 on successful mangle
 *         -1 on error
 *           ENOMEM
 *           ENOENT - there was no username in the parsed line
 **/
int protocol_mangle_username (protocol_input_line *parsed_line, GlobalConf *conf)
{
	char *newusername;
	char *old_domainsep_ptr;

	size_t usernamelen;
	size_t domainlen;
	size_t fulllen;

	// If no username then return with 'error'
	if (parsed_line->username == NULL)
	{
		errno = ENOENT;
		return -1;
	}

	if (conf->unescape) {
		protocol_unescape(parsed_line->username);
	}

	// Is there a domain in here?
	old_domainsep_ptr = strchr (parsed_line->username, '\\');
	if (old_domainsep_ptr == NULL)
	{
		// We've got nothing to do
		return 0;
	}

	// Needs to be the only slash
	if (strchr (old_domainsep_ptr + 1, '\\') != NULL)
	{
		// We cant mangle this
		return 0;
	}

	// Otherwise, allocate new username and swap things around
	fulllen = strlen (parsed_line->username);
	domainlen = old_domainsep_ptr - parsed_line->username;
	usernamelen = fulllen - domainlen - 1;

	// both domain and username fields must be non-zero length
	if (domainlen == 0 || usernamelen == 0)
	{
		return 0;
	}

	// Strip domain if requested
	if (conf->username_remove_domain != NULL) {

		// empty string is a special case: remove any domains
		if (*conf->username_remove_domain == '\0' ) {
			fulllen -= domainlen + 1;
			domainlen = 0;
		} else if (domainlen == conf->username_remove_domain_len &&
				strncmp(parsed_line->username, conf->username_remove_domain, conf->username_remove_domain_len) == 0) {
			fulllen -= domainlen + 1;
			domainlen = 0;
		}
	}

	newusername = malloc (fulllen + 1);
	if (newusername == NULL)
	{
		errno = ENOMEM;
		return -1;
	}

	// Copy the username portion (the part after the slash)
	memcpy (newusername, old_domainsep_ptr + 1, usernamelen);
	if (domainlen != 0) {
		// Now add our new seperator '@'
		newusername[usernamelen] = '@';
		// Now copy in the domain
		memcpy (newusername + usernamelen + 1, parsed_line->username, domainlen);
	}
	newusername[fulllen] = '\0';

	// Swap in the new username and free the old
	free (parsed_line->username);
	parsed_line->username = newusername;

	return 0;
}

/** @brief
 *
 * The iplist should contain the IP to domain mapping.
 * if the username does not look like an email and
 *    the client IP has a matching domain
 * append the @domain to the end of username.
 **/
void protocol_append_domain(struct simple_iphead *iplist,  protocol_input_line *parsed_line) {
	const char *domain;
	int new_len ;
	char *newname;
	char *s=NULL;
	ns_ipaddr nsip;

	if (!iplist)
		return;

	if (!parsed_line->IP)
		return;

	if (!parsed_line->username)
		return;

	s = memchr (parsed_line->IP, '/', strlen(parsed_line->IP));
	if (s) {
		*s = '\0';
	}
	if (inet_pton(AF_INET, parsed_line->IP, &nsip.addr.ip) != 0)
		nsip.family = AF_INET;
	else if (inet_pton(AF_INET6, parsed_line->IP, &nsip.addr.ip6) != 0)
		nsip.family = AF_INET6;
	else
		return;		/* unable to read client IP info */

	if (s)
		*s = '/';

	domain = simple_iplist_find_data(iplist, &nsip);
	if (! domain)
		return;

	new_len = strlen(parsed_line->username) + 1/* @ */ + strlen(domain) + 1/* \0 */;
	newname = malloc(new_len);
	if (! newname)
		return;

	sprintf(newname, "%s@%s", parsed_line->username, domain);
	newname[new_len-1] = '\0';
	free(parsed_line->username);
	parsed_line->username = newname;
	
	return;
}


/** @brief Turn the user name to lowercase
 *
 * @param parsed_line The parsed line with a username you want to mangle
 **/
void protocol_lowercase_username (protocol_input_line *parsed_line)
{
	char *ptr;
	
	for (ptr = parsed_line->username; *ptr && *ptr != '@'; ptr++) {
		*ptr = tolower(*ptr);
	}
}

/** @brief Replace bad characters in username
 *
 *  This replaces all invalid characters in the username of the parsed line
 *  with underscores. The default set of invalid characters can be overridden
 *  by command-line argument.
 *
 * @param conf Configuration structure
 * @param parsed_line The parsed line containing the username whose invalid
 *        characters will be replaced
 *
 * @return 0 on successful replacement of all invalid characters
 *        -1 on error
 *           EINVAL - the parsed line did not contain a username
 **/
int protocol_replace_badchars(GlobalConf *conf, protocol_input_line *parsed_line)
{
	size_t offset;
	regmatch_t match;

	if ( parsed_line->username == NULL ) {
		errno = EINVAL;
		return -1;
	}

	// Start at the beginning of the username
	offset = 0;

	// Loop through the username replacing one badchar at a time, and
	// each time start the search at index +1 past what we just replaced
	while ( regexec(conf->bad_username_chars_preg, parsed_line->username + offset, 1, &match, 0) == 0 ) {

		// Replace the matched character
		*(parsed_line->username + offset + match.rm_so) = '_';

		// rm_so is the offset of the match from the start of our
		// search string -- we sum the offset every time to search past
		// what we've already replaced
		offset += match.rm_so;
	}

	return 0;

}

/********************************************************
 **********  Helper Function Definitions  ***************
 *******************************************************/


/**
 * @brief Allocate and initialize the structure that holds a parsed line of input
 *
 * The structure returned by this function is used by the protocol parsing
 * functions to store the values of the various fields from the input string. It
 * also holds a copy of the unparsed input string.
 *
 * @param input_str_len the length of the unparsed input string
 *
 * @return a new protocol_input_line structure
 *         or NULL if there is a memory allocation error
 */
protocol_input_line *protocol_init (size_t input_str_len)
{
	protocol_input_line *result;
	int i;

	// Allocate a parsed line structure
	result = malloc (sizeof (protocol_input_line) + input_str_len + 1);
	if (result == NULL)
	{
		return NULL;
	}
	// Initialize all pointers
	result->req_id = NULL;
	result->url = NULL;
	result->IP = NULL;
	result->username = NULL;
	result->method = NULL;
	result->group = NULL;
	for (i = 0; i < MAX_PROTOCOLS_OTHER_OPTIONS; i++)
	{
		result->otheroptions[i] = NULL;
	}

	// Set input str to 0 len
	result->line_buffer[0] = '\0';

	return result;
}

/**
 * @brief Find specified protocol in our list of defined protocols
 *
 * @param prot the number of the protocol we're looking for
 *
 * @return the protocol with the specified number
 *         or NULL if we can't find a protocol with that number
 */
const protocol_internal_definition *protocol_find_definition (protocol_select prot)
{
	unsigned int i;

	for (i = 0; i < sizeof (defined_protocols) / sizeof (defined_protocols[0]); i++)
	{
		if (defined_protocols[i].which_protocol == prot)
		{
			return &defined_protocols[i];
		}
	}

	return NULL;
}
