/**
 * @file
 * @brief Loop through all different sizes of request.  Look for problems
 *
 * We want to try all different sizes of url (and the other fields too)
 * And all these in combination with many different values of max request size
 * We will be successful if:
 *  1) redir never segfaults
 *  2) any request <= max request size is properly processed
 *  3) any request > max req size results in either
 *      a) fail close url displayed
 *      b) empty line or the full original url echoed back (fail open)
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "CUnit/Basic.h"
#include "jopen.h"
#include "socketserver.h"

#define BIG_BUFFER_SIZE 65536

static size_t bigreq_server_rcv_count;

/**
 * @brief verify the given generic field is of the form used in this test suite
 *
 * First char should be a dash
 * If total length is more than 10 chars or so then it should have a
 * 5 digit number on the end == total length
 *
 * @returns 0 if field is not of correct form
 **/
static int bigreq_verify_field (const char *field)
{
	int flen = strlen (field);

	return *field == '-' && (flen < 10 || atoi (field + flen - 5) == flen);
}
/**
 * @brief verify the given url is of the form used in this test suite
 *
 * Look for 'http://url' followed by zero or more z
 * followed by a 5 digit number giving field length
 *
 * @returns 0 if url is not of correct form
 **/
static int bigreq_verify_url (const char *url)
{
	const char *prefix = "http://url";
	int plen = strlen (prefix);
	int ulen = strlen (url);
	if (strncmp (url, prefix, plen) != 0)
	{
		return 0;
	}

	if (ulen > 17)
	{
		return atoi (url + ulen - 5) == ulen;
	}
	return 1;
}
/**
 * @brief the fake server calls this for every line received
 *
 * Receive data
 * Confirm that it looks uncorrupted
 * Increment a counter of lines of uncorrupted data received
 **/
static void bigreq_server(struct client_socket* s)
{
	int field = 0;
	char *rxptr;
	char *fieldstart;
	int resplen;

	// To be uncorrupted we need a line feed
	if (strchr (s->rxbuffer, '\n') == NULL)
	{
		// no line feed means just return and hope that
		// our framework does another read and gets us
		// some more data to chew on
		return;
	}

	// Now go through each field and verify contents
	rxptr = s->rxbuffer;
	fieldstart = rxptr;
	while (rxptr - s->rxbuffer < s->rxlen)
	{
		rxptr++;
		// found end of a field?
		if (*rxptr == ' ' || *rxptr == '\n')
		{
			// terminate the field
			*rxptr = '\0';
			// are we on field 0?  If so, verify as url
			if (field == 0)
			{
				if (bigreq_verify_url (fieldstart) == 0)
				{
					s->rxlen=0;
					return;
				}
			}
			// otherwise, verify with generic check
			else if (bigreq_verify_field (fieldstart) == 0)
			{
				s->rxlen=0;
				return;
			}
			fieldstart = rxptr + 1;
			field++;
		}
	}

	// Did we find expected number of fields?
	if (field != 4)
	{
		s->rxlen=0;
		return;
	}

	// Increment success counter
	bigreq_server_rcv_count++;
	// Respond with just the url.  Add a \n
	resplen = strlen (s->rxbuffer);
	s->rxbuffer[resplen] = '\n';
	s->rxbuffer[resplen+1] = '\0';
	socketserver_send(s, s->rxbuffer, resplen + 1);
	s->rxlen=0;
}
/**
 * @brief Add extra chars to some field in the given string
 *
 * Make a string longer by adding some chars to one of the fields
 * field number is base 0
 **/
static int bigreq_addchars (char *strbuf, size_t strbuf_len, int field, int padchars, char thechar)
{
	int len;
	int fieldlen;
	// First find the field where we're gonna add chars
	while (field > 0)
	{
		char *newstr;
		newstr = strchr (strbuf, ' ');
		if (newstr == NULL)
		{
			// there's not another space so this field doesnt exist
			return -1;
		}
		// skip over the space
		newstr += 1;
		// decrement length of total buffer
		strbuf_len -= newstr - strbuf;
		// move strbuf
		strbuf = newstr;
		// decrement field num we're looking for
		field--;
	}
	// Figure out how long the field is already
	fieldlen = 0;
	while (*(strbuf + fieldlen) != '\0' && *(strbuf + fieldlen)!= ' ')
	{
		fieldlen++;
	}
	if (fieldlen == padchars)
	{
		return 0;
	}
	if (fieldlen >= padchars)
	{
		return -1;
	}

	// Then shift everything from past the end of the field to the right
	// to make room for the new chars
	// (Check that we have enough space)
	len = strlen (strbuf + fieldlen);
	if (len + padchars + 1 > (int)strbuf_len)
	{
		return -1;
	}
	memmove (strbuf + padchars, strbuf + fieldlen, len);

	// Then write the new char the specified number of times
	memset (strbuf + fieldlen, thechar, padchars - fieldlen);

	// Finally, if we padded more than 5 chars then
	// go to the last five chars and replace those with
	// a 5 digit count of total length of field
	if (padchars - fieldlen > 5 && padchars < 100000)
	{
		int wasspace;
		wasspace = (*(strbuf + padchars) == ' ') ? 1 : 0;
		sprintf (strbuf + padchars - 5, "%05d", padchars);
		// put the space back if it used to be a space
		*(strbuf + padchars) = wasspace ? ' ' : '\0';
	}

	return 0;
}

/**
 * @brief read one line from the given fd and describe contents
 *
 * Read one line from the given fd and return code to describe contents
 *
 * @returns -1 on error
 *          0 if line is like our test urls
 *          1 if line is empty
 *          2 if line is equal to provided str
 *          3 if none of the above
 **/
static int bigreq_readline_verify (int fd, const char *cmp)
{
	char bigbuffer[BIG_BUFFER_SIZE];
	char *ptr = bigbuffer;
	size_t len;

	len = sizeof (bigbuffer);
	do
	{
		if (read (fd, ptr, 1) < 0)
		{
			fprintf (stderr, "readline_verify: unexpected read error\n");
			return -1;
		}
	} while (len-- > 0 && *ptr++ != '\n');
	// back up one
	if (len == 0)
	{
		fprintf (stderr, "readline_verify: unexpected read len error\n");
		return -1;
	}
	ptr--;
	// strip newline
	*ptr = '\0';

	if (bigreq_verify_url (bigbuffer))
		return 0;
	else if (*bigbuffer == '\0')
		return 1;
	else if (cmp != NULL && strcmp (cmp, bigbuffer) == 0)
		return 2;
	else
	{
		fprintf (stderr, "readline_verify: unexpected response \"%s\"\n", bigbuffer);
		return 3;
	}
}
/**
 * @brief write a range of request lengths to the given file descriptor
 *
 * Will write a variety of requests in a range of request lengths
 * to the provided file descriptor.
 * For each length in the range, this function will vary the length
 * of each request field and send each request seperately
 * Provide the file descriptor to write to and the range of lengths.
 *
 * requests written will be the length specified including the end of line char
 *
 * @returns number of requests written
 **/
static int bigreq_write_req (int fd, int reqsize, int field, const char *basereq)
{
	int basereq_len;
	static char *big_buffer = NULL;
	const size_t big_buffer_len = BIG_BUFFER_SIZE;

	if (big_buffer == NULL) {
		big_buffer = malloc (big_buffer_len);
		if (big_buffer == NULL) {
			return -1;
		}
	}

	basereq_len = strlen (basereq);

	int pad = 0;
	// Print a fresh copy of the base request
	strcpy (big_buffer, basereq);
	// The number of padded chars we need will depend on the field
	switch (field) {
		case 0:
			pad = reqsize - basereq_len + 10 - 1;
			break;
		case 1:
			pad = reqsize - basereq_len + 3 - 1;
			break;
		case 2:
		case 3:
			pad = reqsize - basereq_len + 1 - 1;
	}
	if (bigreq_addchars (big_buffer, big_buffer_len - 1, field, pad, 'z') != 0)
	{
		return -1;
	}

	// Now we have a long request.  add a \n and write it to the FD
	strcpy (big_buffer + reqsize - 1, "\n");
	if (write (fd, big_buffer, reqsize) != reqsize)
	{
		return -2;
	}

	return 0;
}

/**
 * @brief Main function of all tests
 *
 * @param reqsize the max length of requests to send to the redir
 * @param failurl the URL the redir should return when it fails
 *
 * @return 0 on successful test run
 *        -1 on error
 */
static int bigreq_runtest (int reqsize, const char *failurl)
{
	char cmd[1000];
	struct socketserver *server;
	int i;
	int field;
	JFILE *fp;

	// Write cmd
	if (failurl == NULL)
		sprintf (cmd, "../redir -R %d localhost:12345", reqsize);
	else
		sprintf (cmd, "../redir -R %d localhost:12345 -f %s", reqsize, failurl);

	//Create a server
	bigreq_server_rcv_count = 0;
	server = socketserver_create (12345, NULL, bigreq_server, NULL);
	if (server == NULL)
	{
		return -1;
	}
	//Create a redir
	fp = jopen (cmd);
	if (fp == NULL)
	{
		fprintf(stderr, "Could not open ../redir\n");
		goto runtest_cleanup;
	}
	//Write requests ranging from smaller than max req
	//to larger than max req
	//Smaller or equal reqsize should be 'successful'
	//(recived on server and response should be allowed url)
	//Larger should not be succsesful (not receied by server and 
	//linefeed or fail close depending on setting)
	for (i = reqsize - 20; i < reqsize + 50; i++)
	{
		for (field = 0; field < 4; field++)
		{
			// zero this so it can be used again
			bigreq_server_rcv_count = 0;
			if (bigreq_write_req (fp->to_stdin, i, field, "http://url -/- - -") < 0)
			{
				fprintf (stderr, "Error writing request: %d\n", i);
				goto runtest_cleanup;
			}
			if (i <= reqsize)
			{
				// read line and confirm it is a request url
				CU_ASSERT_EQUAL(bigreq_readline_verify (fp->from_stdout, NULL), 0);
			}
			else if (failurl == NULL)
			{
				CU_ASSERT_EQUAL(bigreq_readline_verify (fp->from_stdout, NULL), 1);
			}
			else
			{
				CU_ASSERT_EQUAL(bigreq_readline_verify (fp->from_stdout, failurl), 2);
			}
			//Confirm receipt on server or server got nothing
			CU_ASSERT_EQUAL((int)bigreq_server_rcv_count, (i<=reqsize ? 1 : 0));
		}
	}
	//Close server
	socketserver_signal_shutdown(server);
	socketserver_waitfor_shutdown(server);
	//Close redir
	jclose(fp);

	return 0;
runtest_cleanup:
	if (server != NULL)
	{
		socketserver_signal_shutdown(server);
		socketserver_waitfor_shutdown(server);
	}
	if (fp != NULL)
	{
		jclose (fp);
	}
	return -1;
}

#define DEFINETEST(SIZE)                                                 \
static void bigreq_test_ ##SIZE (void)                                   \
{                                                                        \
	CU_ASSERT_EQUAL(bigreq_runtest (SIZE, NULL), 0);                     \
	CU_ASSERT_EQUAL(bigreq_runtest (SIZE, "http://failtest.com"), 0);    \
}

DEFINETEST(255);
DEFINETEST(256);
DEFINETEST(257);
DEFINETEST(511);
DEFINETEST(512);
DEFINETEST(513);
DEFINETEST(1023);
DEFINETEST(1024);
DEFINETEST(1025);
DEFINETEST(2055);
DEFINETEST(2056);
DEFINETEST(2057);
DEFINETEST(4095);
DEFINETEST(4096);
DEFINETEST(4094);
DEFINETEST(8191);
DEFINETEST(8192);
DEFINETEST(8193);
DEFINETEST(16383);
DEFINETEST(16384);
DEFINETEST(16385);
DEFINETEST(19500);


/**
 * @brief Registers our test cases and test suites with CUnit
 *
 * @return 0 if we registered everything successfully
 *         nonzero otherwise (returns CU_ErrorCode errors)
 */
int bigreq_test_init (void)
{
#define DECLTEST(SIZE) { "max req size " #SIZE, bigreq_test_ ## SIZE }
	CU_TestInfo bigreq_tests[] =
	   {
		DECLTEST(255),
		DECLTEST(256),
		DECLTEST(257),
		DECLTEST(511),
		DECLTEST(512),
		DECLTEST(513),
		DECLTEST(1023),
		DECLTEST(1024),
		DECLTEST(1025),
		DECLTEST(2055),
		DECLTEST(2056),
		DECLTEST(2057),
		DECLTEST(4095),
		DECLTEST(4096),
		DECLTEST(4094),
		DECLTEST(8191),
		DECLTEST(8192),
		DECLTEST(8193),
		DECLTEST(16383),
		DECLTEST(16384),
		DECLTEST(16385),
		DECLTEST(19500),
		CU_TEST_INFO_NULL
	   };

	CU_SuiteInfo bigreq_suites[] =
	   {
		{ "big request with max request test", NULL, NULL, bigreq_tests },
		CU_SUITE_INFO_NULL
	   };

	if (CU_register_suites(bigreq_suites) != CUE_SUCCESS)
	{
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
