/**
 * @file
 * @brief This test runs the redir with various invalid command line params
 *
 *  
 *  For items requiring a choice, do we properly check the choice
 *  For integer params do we check the range?
 *  For integer params do we detect non-number?
 *  For integer params do we detect floating point?
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#include "CUnit/Basic.h"

#include "jopen.h"

/**
 * @brief Helper function attempts to run redir and looks for failure
 *
 * Run the redir with the provided arguments.  Figure out whether we got
 * a usage screen and an error exit code.
 *
 * @params cmdline The arguments string for redir
 *
 * @returns true if redir started and immediately exited with usage
 *          false otherwise
 **/
static int cmdline_gives_usagescreen (const char *cmdline)
{
	JFILE *jp;
	siginfo_t childinfo;
	struct timespec runtime;
	char fullcmdline[2000];

	// Create full cmd line containig program name AND cmd line options
	CU_ASSERT_FATAL(snprintf (fullcmdline, sizeof(fullcmdline), "../redir %s", cmdline) < (int) sizeof (fullcmdline));

	// Run the redir
	jp = jopen (fullcmdline);
	CU_ASSERT_FATAL(jp != NULL);

	// Let it run for 15ms
	//  (It's possible that this isnt long enough to wait for the OS to schedule
	//   the redir app to run and let it run to completion).  We may get
	//   erroneous failures here sometimes
	runtime.tv_sec = 0;
	runtime.tv_nsec =  15000000;
	nanosleep (&runtime, NULL);

	// Now see whether it exited
	childinfo.si_pid = 0;

	if (waitid (P_PID, jp->pid, &childinfo, WEXITED | WNOHANG) == 0 &&
	    childinfo.si_pid != 0)
	{
		char usagescreen[5000];
		int ret;
		// We exited!!
		// Check the return code 
		CU_ASSERT(childinfo.si_status != 0);
		// Get the text written to stdout
		// and look for signs that this is the usage screen
		ret = read (jp->from_stdout, usagescreen, sizeof (usagescreen) - 1);
		CU_ASSERT(ret > 0);
		if (ret > 0)
		{
			usagescreen[ret] = '\0';
			CU_ASSERT(strstr (usagescreen, "Usage: ") != NULL);
		}
		jp->pid = -1;
		jclose (jp);
		return 1;
	}
	
	jclose (jp);
	return 0;
}

/**
 * @brief Test invalid input protocols
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_i (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-i notaprotocol localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-i squid3.01 localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-i squid3.0 localhost:12345"));
}

/**
 * @brief Test invalid output protocols
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_o (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-o notaprotocol localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-o squid3.01 localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-o squid3.0 localhost:12345"));
}

/**
 * @brief Test invalid port
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_P (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-P -500 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-P 70000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-P 0.5 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-P 10hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-P 9956 localhost:12345"));
}

/**
 * @brief Test invalid connections
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_n (void)
{
	// 20000 is the max
	CU_ASSERT(cmdline_gives_usagescreen ("-n 20001 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-n 10.5 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-n -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-n 0 localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-n 50 localhost:12345"));
}

/**
 * @brief Test invalid pattern for username replacement
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_U (void)
{
	// empty pattern 
	CU_ASSERT(cmdline_gives_usagescreen ("-U localhost:12345"));
	// invalid pattern
	CU_ASSERT(cmdline_gives_usagescreen ("-U [^0-9a-zA-Z localhost:12345"));
}

/**
 * @brief Test invalid rewrite
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_u (void)
{
	// this is a boolean option. no invalid values
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-u localhost:12345"));
}

/**
 * @brief Test invalid load balance type
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_b (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-b -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-b 3 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-b 1.5 localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-b 1 localhost:12345"));
}

/**
 * @brief Test invalid fail close host and port number 
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_s (void)
{
	// empty host
	CU_ASSERT(cmdline_gives_usagescreen ("-s localhost:12345"));
	// host with with scheme
	CU_ASSERT(cmdline_gives_usagescreen ("-f ftp://host localhost:12345"));
	// host with path
	CU_ASSERT(cmdline_gives_usagescreen ("-f host/path localhost:12345"));
}

/**
 * @brief Test invalid fail close url
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_f (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-f notaurl localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-f ftp://host localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-f http:// localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-f http://thisisaplausiableurl.ca localhost:12345"));
}

/**
 * @brief Test invalid connect timeout
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_t (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-t 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-t 99.5 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-t -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-t 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-t 10hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-t 900 localhost:12345"));
}

/**
 * @brief Test invalid read timeout
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_r (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-r 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 38.4352 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-r 900 localhost:12345"));
}

/**
 * @brief Test invalid write timeout
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_w (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-w 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-w 123.456 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-w -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-w 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-w 10hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-w 900 localhost:12345"));
}


/**
 * @brief Test invalid reconnect delay
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_d (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-d 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-d 48.2 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-d -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-d hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-d 900 localhost:12345"));
}

/**
 * @brief Test invalid max timeout
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_m (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-m 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-m 2.2 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-m -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-m 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-m 10hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-m 900 localhost:12345"));
}

/**
 * @brief Test invalid max timeouts
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_M (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-r 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 66.7 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-r 10 localhost:12345"));
}

/**
 * @brief Test invalid max request size
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_R (void)
{
	// try a crazy high value
	CU_ASSERT(cmdline_gives_usagescreen ("-r 50000000 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 2.2 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 0 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-r 10hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-r 5000 localhost:12345"));
}

/**
 * @brief Test invalid stderr log
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_e (void)
{
	// This is a boolean option.  No possible invalid values.  Just make
	// sure it works in the positive case
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-e localhost:12345"));
}

/**
 * @brief Test invalid concurrency setting
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_c (void)
{
	// This is a boolean option.  No possible invalid values.  Just make
	// sure it works in the positive case
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-c localhost:12345"));
}

/**
 * @brief Test all valid arguments
 *
 * This test confirms that redir can accept valid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on valid cmdline input
 **/
static void cmdline_valid_args (void)
{
	// This is a boolean option.  No possible invalid values.  Just make
	// sure it works in the positive case
	// TODO: put the -b0 option back in here once load balancing is implemented
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-isquid2.5 -osquid3.1 -P12345 -n5 -u -U[^a-z] -fhttp://failclose.com -sfailclosehost -t100 -r100 -w100 -d100 -m1000 -M10 -R1024 -e -c -L1 localhost:12345"));
}

/**
 * @brief Test invalid log level
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_L (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("-L -1 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-L 1.4 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-L 5 localhost:12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("-L hamburger localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("-L 3 localhost:12345"));
}

/**
 * @brief Test invalid host
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_host (void)
{
	CU_ASSERT(cmdline_gives_usagescreen (":12345"));
	CU_ASSERT(cmdline_gives_usagescreen ("\" :12345\""));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("192.168.100.100:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("localhost:12345"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("localhost"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("netsweeper.com:12345"));
}

/**
 * @brief Test invalid host_port
 *
 * This test confirms that redir can detect invalid cmdline inputs
 *
 * Fails if:
 *   redir fails to exit with FAILURE on any invalid cmdline input
 **/
static void cmdline_inv_host_port (void)
{
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:1hamburger"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost: 3"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:-100"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:10.4"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:99999"));
	CU_ASSERT(cmdline_gives_usagescreen ("localhost:0"));
	CU_ASSERT_FALSE(cmdline_gives_usagescreen ("localhost:12345"));
}


int cmdline_test_init (void)
{
	CU_TestInfo invalid_arg_tests[] =
	   {
		{ "invalid input protocols", cmdline_inv_i },
		{ "invalid output protocols", cmdline_inv_o },
		{ "invalid port", cmdline_inv_P },
		{ "invalid connections", cmdline_inv_n },
		{ "invalid rewrite", cmdline_inv_u },
		{ "invalid bad char pattern", cmdline_inv_U },
		//{ "invalid load balance type", cmdline_inv_b },  // load balancing not implemented
		{ "invalid fail close url", cmdline_inv_f },
		{ "invalid fail close host", cmdline_inv_s },
		{ "invalid connect timeout", cmdline_inv_t },
		{ "invalid read timeout", cmdline_inv_r },
		{ "invalid write timeout", cmdline_inv_w },
		{ "invalid reconnect delay", cmdline_inv_d },
		{ "invalid max timeout", cmdline_inv_m },
		{ "invalid max timeouts", cmdline_inv_M },
		{ "invalid max request size", cmdline_inv_R },
		{ "invalid stderr log", cmdline_inv_e },
		{ "invalid server concurrent", cmdline_inv_c },
		{ "invalid log level", cmdline_inv_L },
		{ "invalid host", cmdline_inv_host },
		{ "invalid host_port", cmdline_inv_host_port },
		{ "valid arguments", cmdline_valid_args },
		CU_TEST_INFO_NULL
	   };

	CU_SuiteInfo cmdline_suites[] =
	   {
		{ "command line invalid options test", NULL, NULL, invalid_arg_tests },
		CU_SUITE_INFO_NULL
	   };

	if (CU_register_suites(cmdline_suites) != CUE_SUCCESS)
	{
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
