/**
 * @file
 * @brief Test the concurrency features of the redirector
 *
 * The following tests the concurrency features of the redirector.  This feature
 * allows many requests to be queued at any given time and processed as fast as
 * possible.
 *
 * The INPUT and OUTPUT order will not match in this case.
 *
 * When concurrency is used each request in INPUT must be flagged with some type
 * of unique identifier.  This test case does not test this, only the stability
 * and processing on concurrent requests.
 *
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <sys/select.h>
#include <ctype.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

struct conninfo {
	int total_count;
	int connections;
	int* counters;
};
static struct conninfo* conninfo_init()
{
	struct conninfo* ci;
	ci = malloc(sizeof(struct conninfo));
	ci->total_count=0;
	ci->connections=0;
	ci->counters=NULL;
	return ci;
}
static void conninfo_free(struct conninfo* ci)
{
	if ( ci->counters ) {
		free(ci->counters);
	}
	free(ci);
}

static void connect_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;

	// our current connection id in the counters array
	s->tag = (void*)ci->connections;
	// increment, then reallocate
	ci->connections++;
	ci->counters =  realloc(ci->counters, sizeof(int) * ci->connections);
	ci->counters[(int)s->tag]=0;
	printf("%p: Connected number %d\n", s, (int)s->tag);
}

/**
 * @brief This call back will increment our conninfo information and allow the request
 */
static void recv_allow_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;

		ci->total_count++;
		ci->counters[(int)s->tag]++;
	}
}

/**
 * @brief This call back simulates a server that doesn't recognize message ID
 */
static void recv_allow_noID_callback(struct client_socket* s)
{
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		if (isdigit(s->rxbuffer[0])) {
			socketserver_send(s, "\n", 1);
			s->rxlen=0;
		} else {
			socketserver_send(s, s->rxbuffer, s->rxlen);
			s->rxlen=0;
		}
	}
}

/**
 * @brief This call back simulates a server that recognizes message ID
 */
static void recv_allow_ID_callback(struct client_socket* s)
{
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}

/**
 * @brief This will do the same as recv_allow_callback but will also disconnect
 */
static void recv_disconnect_callback(struct client_socket* s)
{
	// do the allow callback
	recv_allow_callback(s);

	// disconnect
	socketserver_connection_shutdown(s);
}

/**
 * @brief This will simply not send anything, noop
 */
static void recv_noop_callback(struct client_socket* s)
{
	// do nothing... but consume the data
	s->rxlen=0;
}

/**
 * @brief This sends some garbage data back with no \n
 */
static void recv_garbage_callback(struct client_socket* s)
{
	// send some garbage
	socketserver_send(s, "welcome, this is the farm house", 10);
	// consume
	s->rxlen=0;
}

/**
 * @brief This callback does random errors
 */
static void recv_randomerrors_callback(struct client_socket* s)
{
	static int nexterror=0;

	// every 20 requests
	if ( (nexterror % 20) == 0 ) {
		// consume the data
		s->rxlen=0;
	}
	// every 10 requests
	else if ( (nexterror % 10) == 0 ) {
		// allow
		recv_allow_callback(s);
		// disconnect
		socketserver_connection_shutdown(s);
	}
	// otherwise, allow
	else {
		// allow
		recv_allow_callback(s);
	}

	nexterror++;
}

static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

/**
 * @brief Create two servers and send 100 requests, then read 100 requests
 *
 * This is concurrent, so we basically generate 100 requests all at once
 * We when process the responses all at once...
 *
 * This test creates two servers, so the concurrent requests are segmented
 * between the two.  You should notice that when you run this test case
 * that the distribution is no longer 50/50 between the two servers.
 *
 */
static void concurrent_simple_twoserver_test()
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[1025];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_allow_callback, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl -i squid2.6 -o squid2.6");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return;
	}

	// Perform 100 tests two each server
	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", i, i);
		printf("send %s\n", sendbuffer);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
	}

	int okay[100];
	for(i=0; i<100; i++) {
		okay[i]=0;
	}


	// Read chunks of data and process them
	// Note: one read may read many requests, so we loop through process
	// the requests
	char* end;
	int count=0;
	buffer[0]='\0';
	while ( count < 100 ) {
		printf("Reading Data %d\n", count);

		// Read and terminate string
		end = buffer+strlen(buffer);
		ret=read(fp->from_stdout, buffer +strlen(buffer), 1024 - strlen(buffer));
		printf("Read %d bytes\n", ret);
		if ( ret >= 0 ) {
			*(end+ret)='\0';
		}

		// each response is \n terminated
		// so we process each one
		while ( (end = strchr(buffer, '\n')) ) {
			*end = '\0';
			end++;

			ptr=strchr(buffer, ' ');
			*ptr='\0';
			okay[atoi(buffer)]=1;

			// move the buffer
			count++;
			memmove(buffer, end, strlen(end)+1);
		}

	}

	int failed = 0;
	for(i=0; i<100; i++) {
		if ( okay[i] != 1 ) {
			failed++;
		}
	}

	printf("Failed Requests %d\n", failed);
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);


	// we did not get all the responses
	if ( failed != 0 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("We did not recv all the responses");
		return;
	}

	// We processed to many requests on a single server
	if ( ci1->total_count + ci2->total_count != 100 ) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("One server instance processed more then 80 requests.  Distribution not very good");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}

/**
 * @brief Test message ID support
 *
 * The server ID can be processed on server and in redir.
 * This test tries all for combinations:
 * - redir doesn't process ID and server doesn't process ID
 * - redir processes ID but server doesn't process ID
 * - redir doesn't process ID but server processes ID
 * - redir processes ID and server can process ID
 *
 * The fisrt case should fail and sender shoud get a message without ID
 * other three cases should produce correct responces with message ID 
 */
static void concurrent_server_concurrency_test()
{
	struct socketserver* svr1;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;

	// Test 1 : neither redir nor server support concurrency
	// Create a server
	svr1 = socketserver_create(12345, connect_callback, recv_allow_noID_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir -c localhost:12345 -f http://failcloseurl -i squid2.6 -o squid2.6");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 10 tests for this server
	for(i=0; i<10; i++) {
		sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", i, i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate buffer at linefeed
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strlen(buffer) > 0 ) {
			fprintf(stderr, "ret = %d buffer %s when just '\\n' expected\n", ret, buffer);
			CU_FAIL("We did not get the response we expected on sub-test 1");
			socketserver_signal_shutdown(svr1);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr1);
	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);


	// Test 2 : redir supports concurrency but server doesn't
	// Create a server
	svr1 = socketserver_create(12345, connect_callback, recv_allow_noID_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 -f http://failcloseurl -i squid2.6 -o squid2.6");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 10 tests for this server
	for(i=0; i<10; i++) {
		sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", i, i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at linefeed
		ptr=strchr(sendbuffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}
		// truncate buffer at linefeed
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected on sub-test 2");
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr1);
	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);

	// Test 3 : redir rely on server support and server supports concurrency
	// Create a server
	svr1 = socketserver_create(12345, connect_callback, recv_allow_ID_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir -c localhost:12345 -f http://failcloseurl -i squid2.6 -o squid2.6");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 10 tests for this server
	for(i=0; i<10; i++) {
		sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", i, i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at linefeed
		ptr=strchr(sendbuffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}
		// truncate buffer at linefeed
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected on sub-test 3");
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr1);
	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);

	// Test 4 : both redir and server support concurrency
	// Create a server
	svr1 = socketserver_create(12345, connect_callback, recv_allow_ID_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 -f http://failcloseurl -i squid2.6 -o squid2.6");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return;
	}

	// Perform 10 tests for this server
	for(i=0; i<10; i++) {
		sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", i, i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at linefeed
		ptr=strchr(sendbuffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}
		// truncate buffer at linefeed
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr1);
	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	jclose(fp);


	CU_PASS("The Concurrency Support test has passed");
	return;
}

/**
 * @brief This is our framework function for creating super crazy tests
 *
 */
static void concurrent_framework(SOCKETSERVER_CALLBACK svr1_callback, SOCKETSERVER_CALLBACK svr2_callback)
{
	struct socketserver* svr1=NULL;
	struct socketserver* svr2=NULL;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[1025];
	int ret;
	char* ptr;
	struct conninfo* ci1=NULL;
	struct conninfo* ci2=NULL;
#define LONG_MAX_REQUESTS 100000
	int okay[LONG_MAX_REQUESTS];
	// Perform LONG_MAX_REQUESTS tests two each server
	int sent;
	int sendcount=0;
	int recvcount=0;
	int nfds;
	fd_set readers;
	fd_set writers;
	char* end;
	int failed = 0;
	int total_count = 0;

	buffer[0]='\0';
	for(i=0; i<LONG_MAX_REQUESTS; i++) {
		okay[i]=0;
	}


	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, svr1_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);

	// create the second server if we require it
	if ( svr2_callback ) {
		svr2 = socketserver_create(12346, connect_callback, svr2_callback, disconnect_callback);
		svr2->tag = conninfo_init();
		socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);
		fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl -i squid2.6 -o squid2.6");
	}
	else {
		// If we only have one server
		// open 2 connections to this one server
		fp = jopen("../redir localhost:12345 -n 2 -f http://failcloseurl -i squid2.6 -o squid2.6");
	}

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		if ( svr2_callback ) {
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr2);
		}
		return;
	}

	do {
		// we select to find out if we can read/write to the redir
		// when we can write we write (if we need to)
		// and when we can read we read... (because we always need to)
		FD_ZERO(&readers);
		FD_SET(fp->from_stdout, &readers);
		FD_ZERO(&writers);
		nfds = fp->from_stdout;
		if ( sendcount < LONG_MAX_REQUESTS ) {
			FD_SET(fp->to_stdin, &writers);
			if ( fp->to_stdin > nfds ) {
				nfds=fp->to_stdin;
			}
		}

		ret = select(nfds+1, &readers, &writers, NULL, 0);
		if ( ret <= 0 ) {
			socketserver_signal_shutdown(svr1);
			socketserver_waitfor_shutdown(svr1);
			if ( svr2_callback ) {
				socketserver_signal_shutdown(svr2);
				socketserver_waitfor_shutdown(svr2);
			}
			jclose(fp);
			CU_FAIL("Select failure, may indicate child death of redir");
			return;
		}

		// we can send something
		if ( FD_ISSET(fp->to_stdin, &writers) && sendcount < LONG_MAX_REQUESTS)
		{
			// Chunking send
			sprintf(sendbuffer, "%d http://google.ca/%d -/- - - -\n", sendcount, sendcount);
			/*
			printf("SENDING: %s\n", sendbuffer);
			printf("--------------------------------------------------------------------------------\n");
			*/
			sent=0;
			do {
				ret = write(fp->to_stdin, sendbuffer+sent, strlen(sendbuffer) - sent);
				if ( ret > 0 ) {
					sent+=ret;
				}
				else {
					// error out
					socketserver_signal_shutdown(svr1);
					socketserver_waitfor_shutdown(svr1);
					if ( svr2_callback ) {
						socketserver_signal_shutdown(svr2);
						socketserver_waitfor_shutdown(svr2);
					}
					CU_FAIL("Could not send request");
					return;
				}
			} while ( sent != (int)strlen(sendbuffer) );
			sendcount++;
		}


		// Check to see if we can recv a little
		if ( FD_ISSET(fp->from_stdout, &readers) && recvcount < LONG_MAX_REQUESTS) {
			// Read chunks of data and process them
			// Note: one read may read many requests, so we loop through process
			// the requests

			// Read and terminate string
			end = buffer+strlen(buffer);
			ret=read(fp->from_stdout, buffer +strlen(buffer), 1024 - strlen(buffer));
			if ( ret >= 0 ) {
				*(end+ret)='\0';
			}

			// each response is \n terminated
			// so we process each one
			while ( (end = strchr(buffer, '\n')) ) {
				*end = '\0';
				end++;

				printf("GOT '%s'\n", buffer);
				ptr=strchr(buffer, ' ');
				*ptr='\0';
				//printf("GOT REQUEST %d\n", atoi(buffer));
				okay[atoi(buffer)]=1;

				// move the buffer
				recvcount++;
				memmove(buffer, end, strlen(end)+1);
			}
		}
	} while ( recvcount < LONG_MAX_REQUESTS);

	for(i=0; i<LONG_MAX_REQUESTS; i++) {
		if ( okay[i] != 1 ) {
			failed++;
		}
	}

	printf("Failed Requests %d\n", failed);
	

	socketserver_signal_shutdown(svr1);
	if ( svr2_callback ) {
		socketserver_signal_shutdown(svr2);
	}

	ci1=svr1->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	if ( svr2_callback ) {
		ci2=svr2->tag;
		printf("Server 2 Processed: %d requests\n", ci2->total_count);
	}


	// we did not get all the responses
	if ( failed != 0 ) {
		socketserver_waitfor_shutdown(svr1);
		if ( svr2_callback ) {
			socketserver_waitfor_shutdown(svr2);
		}
		jclose(fp);
		CU_FAIL("We did not recv all the responses");
		return;
	}

	if ( svr2_callback ) {
		total_count = ci1->total_count + ci2->total_count;
	}
	else {
		total_count = ci1->total_count;
	}

	// We processed to many requests on a single server
	if ( total_count != LONG_MAX_REQUESTS) {
		socketserver_waitfor_shutdown(svr1);
		if ( svr2_callback ) {
			socketserver_waitfor_shutdown(svr2);
		}
		jclose(fp);
		CU_FAIL("We did not process the entire number of requests properly.  Both servers did not see all the requests.");
		return;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	if ( svr2_callback ) {
		socketserver_waitfor_shutdown(svr2);
	}
	jclose(fp);

	CU_PASS("The Multiple Server Redir test has passed");
	return;
}


/**
 * @brief This is a much more robust concurrent test that sends and recvs concurrent requests at once
 *
 * This test is a all okay everything is fine and dandy test case
 *
 */
static void concurrent_2servers_okay_test()
{
	concurrent_framework(recv_allow_callback, recv_allow_callback);
}

/**
 * @brief This tests one good server, one unstable server in concurrent mode
 *
 */
static void concurrent_2servers_1good_1disconnect_test()
{
	concurrent_framework(recv_allow_callback, recv_disconnect_callback);
}

/**
 * @brief This tests one good server, and one that just never sends any data
 *
 */
static void concurrent_2servers_1good_1noop_test()
{
	concurrent_framework(recv_allow_callback, recv_noop_callback);
}
/**
 * @brief This tests one good server, and one that just never sends crap data
 *
 */
static void concurrent_2servers_1good_1garbage_test()
{
	concurrent_framework(recv_allow_callback, recv_garbage_callback);
}
/**
 * @brief One server is created for this test but we randomly generate errors
 */
static void concurrent_1server_randomerrors_test()
{
	concurrent_framework(recv_randomerrors_callback, NULL);
}
/**
 * @brief Two servers are created for this test but we randomly generate errors
 */
static void concurrent_2servers_randomerrors_test()
{
	concurrent_framework(recv_randomerrors_callback, recv_randomerrors_callback);
}


int concurrent_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("Conncurrent Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Send 100 requests, recv 100 test", concurrent_simple_twoserver_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Concurrent send/recv to good servers", concurrent_2servers_okay_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Concurrent send/recv 1 good 1 disconnecting server", concurrent_2servers_1good_1disconnect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Concurrent send/recv 1 good 1 noop server", concurrent_2servers_1good_1noop_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Concurrent send/recv 1 good 1 garbage server", concurrent_2servers_1good_1garbage_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// One server with 2 connections, random errors
	if ( (NULL == CU_add_test(pSuite, "Concurrent 1 server 2conns random errors", concurrent_1server_randomerrors_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	// Two servers, random errors
	if ( (NULL == CU_add_test(pSuite, "Concurrent 2 server random errors", concurrent_2servers_randomerrors_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance
	if ( (NULL == CU_add_test(pSuite, "Message ID support test", concurrent_server_concurrency_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	return 0;
}
