/**
 * @file
 * @brief This is the base for all unit tests
 *
 * Open the redir
 * Send a request
 * Make sure we get the exact same thing back
 * Repeat a few times
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

static void connect_callback(struct client_socket* s)
{
	printf("%p: Connected\n", s);
}
static void recv_callback(struct client_socket* s)
{
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}

	// Shutdown this connection
	socketserver_connection_shutdown(s);
}
static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

static void deadserver_disconnect_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
/*
	signal(SIGTERM, stop);
	signal(SIGQUIT, stop);
	signal(SIGINT, stop);
	signal(SIGPIPE,SIG_IGN);
	*/

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl -L3");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<1000; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void recv_nodata_callback(struct client_socket* s)
{
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// Consumed...
		s->rxlen=0;
	}
}
static void recv_noreturn_callback(struct client_socket* s)
{
	char* ptr;
	printf("%p: Data %s", s, s->rxbuffer);
	if ( (ptr=strchr(s->rxbuffer, '\n' )) ) {
		*ptr='\0';
		socketserver_send(s, s->rxbuffer, strlen(s->rxbuffer));
		// Consumed...
		s->rxlen=0;
	}
}
static int deadserver_deadserver_onrequest=0;
static void recv_deadserver_onrequest_callback(struct client_socket* s)
{
	char* ptr;
	printf("%p: Data %s", s, s->rxbuffer);
	if ( (ptr=strchr(s->rxbuffer, '\n' )) ) {
		if ( deadserver_deadserver_onrequest ) {
			*ptr='\0';
		}
		socketserver_send(s, s->rxbuffer, strlen(s->rxbuffer));
		// Consumed...
		s->rxlen=0;
	}
}

static void deadserver_nodata_inresponse_failclose_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
/*
	signal(SIGTERM, stop);
	signal(SIGQUIT, stop);
	signal(SIGINT, stop);
	signal(SIGPIPE,SIG_IGN);
	*/

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_nodata_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, "http://failcloseurl") ) {
			fprintf(stderr, "ret = %d buffer %s != http://failcloseurl\n", ret, buffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void deadserver_noreturn_inresponse_failclose_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
/*
	signal(SIGTERM, stop);
	signal(SIGQUIT, stop);
	signal(SIGINT, stop);
	signal(SIGPIPE,SIG_IGN);
	*/

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_noreturn_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, "http://failcloseurl") ) {
			fprintf(stderr, "ret = %d buffer %s != http://failcloseurl\n", ret, buffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void deadserver_deadserver_failopen_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
/*
	signal(SIGTERM, stop);
	signal(SIGQUIT, stop);
	signal(SIGINT, stop);
	signal(SIGPIPE,SIG_IGN);
	*/

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_noreturn_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void deadserver_deadserver_1request_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
/*
	signal(SIGTERM, stop);
	signal(SIGQUIT, stop);
	signal(SIGINT, stop);
	signal(SIGPIPE,SIG_IGN);
	*/

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_deadserver_onrequest_callback, disconnect_callback);
	deadserver_deadserver_onrequest=1;

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl -m5000");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( i != 0 ) {
			if ( strcmp(buffer, sendbuffer) ) {
				fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
				CU_FAIL("We did not get the response we expected.");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
		}
		else {
			// We should fail close
			if ( strcmp(buffer, "http://failcloseurl") ) {
				fprintf(stderr, "ret = %d buffer %s != http://failcloseurl\n", ret, buffer);
				CU_FAIL("We did not get the response we expected.");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
			// delay for 6 seconds
			printf("Waiting for 6 seconds for redir to recover from dead server\n");
			deadserver_deadserver_onrequest=0;
			sleep(6);
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}



int deadserver_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test4
	pSuite = CU_add_suite("Dead Server Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Disconnect after request test", deadserver_disconnect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "No data fail close test", deadserver_nodata_inresponse_failclose_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Make sure if we get mangled data, we fail close
	if ( (NULL == CU_add_test(pSuite, "No line return fail close test", deadserver_noreturn_inresponse_failclose_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// This just makes sure fail open works as well
	if ( (NULL == CU_add_test(pSuite, "No data fail open test", deadserver_deadserver_failopen_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Make sure we somehow recover from no data
	if ( (NULL == CU_add_test(pSuite, "No data recovery test", deadserver_deadserver_1request_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
