/**
 * @file
 * @brief Tests to make sure the redirector can fail open or fail close properly
 *
 * Run redir with nsp server
 *
 * Keep running redir
 * Run nsp server
 * Shut down nsp server
 * Restart nsp server
 * response should be fail open or fail close respectively
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include "jopen.h"

#include "CUnit/Basic.h"

#include "socketserver.h"
static void connect_callback(struct client_socket* s)
{
	printf("%p: Connected\n", s);
}
static void recv_callback(struct client_socket* s)
{
	int len;
	printf("%p: Data %s", s, s->rxbuffer);

	if ( strchr(s->rxbuffer, '\n')) {
		len = 0;
		while ( s->rxlen - len > 0 ) {
			socketserver_send(s, s->rxbuffer+len, 1);
			len++;
		}
		s->rxlen=0;
	}
}
static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

void redir_fail_open_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input and output
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i); 
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail open");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}

	// shut down nsp
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// fail open
	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail open");
			jclose(fp);
			return;
		}
	}

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);
	usleep(10000000);

	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i); 
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail open");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	CU_PASS("We have passed this test - redir fail open");
	jclose(fp);
	return;
}

void redir_fail_close_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	char expbuffer[256];
	int ret;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);

	// squid 2.5 input and output
	fp = jopen("../redir localhost -i squid2.5 -o squid2.5 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i); 
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail close");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}

	// shut down nsp
	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);

	// fail close 
	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://failcloseurl\n");
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail close");
			jclose(fp);
			return;
		}
	}

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_callback, disconnect_callback);
	usleep(10000000);

	for(i=0; i<10; i++) {
		// simulate the buffer sent from squid
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		// the buffer redir expects to send
		sprintf(expbuffer, "http://google.ca/%d -/- - -\n", i); 
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

		int len=0;
		buffer[0]='\0';
		while ( strchr(buffer, '\n') == NULL ) {
			ret=read(fp->from_stdout, buffer+len, 1024);
			if ( ret > 0 ) {
				len+=ret;
				buffer[len]='\0';
			}
		}

		if ( strcmp(buffer, expbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != expect buffer %s\n", ret, buffer, expbuffer);
			CU_FAIL("We did not get the redirector output we expected for test - redir fail close");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	CU_PASS("We have passed this test - redir fail close");
	jclose(fp);
	return;
}

int redir_fail_open_init()
{
	CU_pSuite pSuite = NULL;

	pSuite = CU_add_suite("Redirector Fail Open Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Redire Fail Open Test\n", redir_fail_open_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	if ( (NULL == CU_add_test(pSuite, "Redire Fail Close Test\n", redir_fail_close_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	return 0;
}
