/**
 * @file
 * @brief This files tests the redir in server cases where a server constantly changes state
 *
 * The goal of this test case is to determine acceptable performance characteristics when
 * a server is constantly changing state.  This may mean it comes up and down frequently, or
 * even that connection to this server are often interrupted, disconnected, etc.
 *
 * The following logic is used in these tests
 *
 * Get a baseline of performance in requests/second we can process to one server
 * that is always up and available.
 *
 * In theory even with a flapping server should should be able to maintain the request
 * rate and experience no fail open fail close situations.
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>
#include <sys/time.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

#define BASELINE_REQUESTS_TOTAL 10000

struct conninfo {
	int total_count;
	int connections;
	int* counters;
};
static struct conninfo* conninfo_init()
{
	struct conninfo* ci;
	ci = malloc(sizeof(struct conninfo));
	ci->total_count=0;
	ci->connections=0;
	ci->counters=NULL;
	return ci;
}
static void conninfo_free(struct conninfo* ci)
{
	if ( ci->counters ) {
		free(ci->counters);
	}
	free(ci);
}

static void connect_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;

	// our current connection id in the counters array
	s->tag = (void*)ci->connections;
	// increment, then reallocate
	ci->connections++;
	ci->counters =  realloc(ci->counters, sizeof(int) * ci->connections);
	ci->counters[(int)s->tag]=0;
	printf("%p: Connected number %d\n", s, (int)s->tag);
}

/**
 * @brief This call back will increment our conninfo information and allow the request
 */
static void recv_allow_callback(struct client_socket* s)
{
	struct conninfo* ci = (struct conninfo*)s->srv->tag;
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;

		ci->total_count++;
		ci->counters[(int)s->tag]++;
	}
}
/**
 * @brief Call the allow then disconnect
 */
static void recv_disconnect_callback(struct client_socket* s)
{
	recv_allow_callback(s);
	socketserver_connection_shutdown(s);
}

/**
 * @brief Consume all data and do nothing
 */
static void recv_noop_callback(struct client_socket* s)
{
	s->rxlen=0;
}

static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}


static double flapping_requestrate(SOCKETSERVER_CALLBACK recv_func)
{
	struct socketserver* svr1;
	struct socketserver* svr2;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;
	struct conninfo* ci1;
	struct conninfo* ci2;
	struct timeval start;
	struct timeval end;
	struct timeval final;
	double request_rate;

	// Create two servers and always allow requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);
	svr1->tag = conninfo_init();
	socketserver_tagfree_func(svr1, (TAGFREE_CALLBACK)conninfo_free);
	svr2 = socketserver_create(12346, connect_callback, recv_func, disconnect_callback);
	svr2->tag = conninfo_init();
	socketserver_tagfree_func(svr2, (TAGFREE_CALLBACK)conninfo_free);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost:12345 localhost:12346 -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_signal_shutdown(svr2);
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		return -1;
	}

	// The baseline must be over a large number of requests
	printf("Performing %d requests for baseline\n", BASELINE_REQUESTS_TOTAL);
	gettimeofday(&start, NULL);
	for(i=0; i<BASELINE_REQUESTS_TOTAL; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, 1024);
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr1);
			socketserver_signal_shutdown(svr2);
			socketserver_waitfor_shutdown(svr1);
			socketserver_waitfor_shutdown(svr2);
			jclose(fp);
			return -1;
		}
	}
	gettimeofday(&end, NULL);


	// Calculate how long it took
	timersub(&end, &start, &final);

	request_rate = (double)BASELINE_REQUESTS_TOTAL / ((double)final.tv_sec + (double) final.tv_usec / (double) 1000000.0);
	printf("Request Rate: %f\n", request_rate);
	

	socketserver_signal_shutdown(svr1);
	socketserver_signal_shutdown(svr2);

	ci1=svr1->tag;
	ci2=svr2->tag;
	printf("Server 1 Processed: %d requests\n", ci1->total_count);
	printf("Server 2 Processed: %d requests\n", ci2->total_count);

	// We processed to many requests on a single server
	if ( ci1->total_count + ci2->total_count != BASELINE_REQUESTS_TOTAL) {
		socketserver_waitfor_shutdown(svr1);
		socketserver_waitfor_shutdown(svr2);
		jclose(fp);
		CU_FAIL("One server instance processed more then 80 requests.  Distribution not very good");
		return -1;
	}

	// We have passed this test
	socketserver_waitfor_shutdown(svr1);
	socketserver_waitfor_shutdown(svr2);
	jclose(fp);

	CU_PASS("The Multiple Server Redir test has passed");
	return request_rate;
}


/**
 * @brief Calculate the difference between two good servers, and one good/one flapping
 *
 * Generate a dual server setup
 * Calculate the request rate
 *
 * Generate a dual server, with one flapping
 * Calculate the request rate
 *
 * Compare the two request rates and take a peek at the results
 *
 *
 */
static void flapping_requestrate_test(SOCKETSERVER_CALLBACK callback)
{
	double rr_diff;
	double rr_percent;
	double rr_good;
	double rr_flapping;

	rr_good = flapping_requestrate(recv_allow_callback);
	if ( rr_good == -1 ) {
		return;
	}
	rr_flapping = flapping_requestrate(callback);
	if ( rr_flapping == -1 ) {
		return;
	}

	printf("THE DISCONNECT AFTER REQUEST TEST CAN BE 50%% slower or faster\n");
	printf("Good: %f Flapping: %f\n", rr_good, rr_flapping);

	// Calculate the difference
	rr_diff = rr_good - rr_flapping;
	printf("Diff: %f\n", rr_diff);

	// The flapping test should have been slower... but with disconnects
	// it maybe faster since this is a quick recovery time
	if ( rr_diff < 0 ) {
		rr_diff *= -1;
	}

	// Fail the test if we are 10% slower with a flapping server
	// There will be an expected 50% slowdown, which is caused by 
	// the decrement from 2 to 1 of working servers number.
	rr_percent = (rr_diff / rr_good) * 100.0;
	printf("Percentage Slower: %f\n", rr_percent);
	if ( rr_percent > 50.0 + 10.0 ) {
		CU_FAIL("One good server and one flapping server are 60% slower than two good servers");
	}

}


/**
 * @brief Calculate the difference between two good servers, and one good/one flapping
 *
 * Generate a dual server setup
 * Calculate the request rate
 *
 * Generate a dual server, with one flapping
 * Calculate the request rate
 *
 * Compare the two request rates and take a peek at the results
 *
 * Uses the recv_disconnect_callback
 *
 */
static void flapping_requestrate_disconnect_test()
{
	flapping_requestrate_test(recv_disconnect_callback);
}

/**
 * @brief Calculate the difference between two good servers, and one good/one flapping
 *
 * Generate a dual server setup
 * Calculate the request rate
 *
 * Generate a dual server, with one flapping
 * Calculate the request rate
 *
 * Compare the two request rates and take a peek at the results
 *
 * Uses the recv_noop_callback
 *
 */
static void flapping_requestrate_noop_test()
{
	flapping_requestrate_test(recv_noop_callback);
}


int flapping_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("Flapping Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance properly
	if ( (NULL == CU_add_test(pSuite, "Flapping Server Request Rate Test Disconnecting", flapping_requestrate_disconnect_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Two servers, load balance properly
	if ( (NULL == CU_add_test(pSuite, "Flapping Server Request Rate Test No Response", flapping_requestrate_noop_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
