/**
 * @file
 * @brief Simple replacement for popen that allowes for stdin, stdout and stderr pipes
 *
 *
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include "jopen.h"

/**
 * @brief Create the pipe and execute the application
 * @param jp The app definition we want to run, already setup
 */
static void jopen_handler(JFILE* jp, int stdin_pipe[2], int stdout_pipe[2], int stderr_pipe[2])
{
	if ( jp== NULL ) {
		return;
	}

	dup2(stdin_pipe[0], 0); // input
	close(stdin_pipe[0]);
	dup2(stdout_pipe[1], 1); // output
	close(stdout_pipe[1]);
	dup2(stderr_pipe[1], 2); // output
	close(stderr_pipe[1]);

	// stdin and std out are redirected
	// we now use the exec function call to run the external module
	
	execv(jp->argv[0], jp->argv);
}


/**
 * @brief open a program and execute it
 *
 * @param program The program with arguments
 * @return JFILE* The reference to the program
 */
JFILE* jopen(char* program)
{
	int stdout_pipe[2] = {-1,-1};
	int stdin_pipe[2] = {-1,-1};
	int stderr_pipe[2] = {-1,-1};
	JFILE* jp;

	jp = malloc(sizeof(JFILE));
	if ( ! jp ) {
		return NULL;
	}

	jp->program=NULL;
	jp->to_stdin = -1;
	jp->from_stdout = -1;
	jp->from_stderr = -1;

	// Create the pipes
	if ( pipe(stdout_pipe) == -1 ) {
		goto jopen_error;
	}
	if ( pipe(stderr_pipe) == -1 ) {
		goto jopen_error;
	}
	if ( pipe(stdin_pipe) == -1 ) {
		goto jopen_error;
	}

	jp->program = strdup(program);
	char **ap;
	for (ap = jp->argv; (*ap = strsep(&jp->program, " \t")) != NULL;) {
		if (**ap != '\0') {
			if (++ap >= &jp->argv[200]) {
				break;
			}
		}
	}

	jp->pid = fork();
	if ( jp->pid == -1 ) {
		goto jopen_error;
	}
	else if ( jp->pid == 0 ) {
		// I'm the child process
		// close the parts of the pipe I wont use
		close(stdin_pipe[1]);
		close(stdout_pipe[0]);
		close(stderr_pipe[0]);
		jopen_handler(jp, stdin_pipe, stdout_pipe, stderr_pipe);
		exit(0);
	}
	else {
		// I'm the parent process
		// close the parts of the pipe I wont use
		close(stdin_pipe[0]);
		close(stdout_pipe[1]);
		close(stderr_pipe[1]);

		// Remember the pipe file handles using
		// descriptive names
		jp->to_stdin = stdin_pipe[1];
		jp->from_stdout = stdout_pipe[0];
		jp->from_stderr = stderr_pipe[0];
	}

	return jp;
jopen_error:
	jclose(jp);
#define CCLOSE(FD) if ((FD) != -1) { close (FD); }
	CCLOSE(stdout_pipe[0]);
	CCLOSE(stdout_pipe[1]);
	CCLOSE(stdin_pipe[0]);
	CCLOSE(stdin_pipe[1]);
	CCLOSE(stderr_pipe[0]);
	CCLOSE(stderr_pipe[1]);
#undef CCLOSE
	return NULL;
}

/**
 * @brief Close/terminate the program
 *
 * @param jp The program we have up and running
 */
void jclose(JFILE* jp)
{
	int i, maxtry = 50;
	siginfo_t childinfo;
	if ( jp->pid > 0 ) {
		kill(jp->pid, SIGTERM);

		childinfo.si_pid = 0;	
		for(i = 0; i < maxtry; i++) {
			if(waitid(P_PID, jp->pid, &childinfo, WEXITED | WNOHANG) == 0 && childinfo.si_pid != 0) {
				printf("Redir process in test was shutted down after waiting %d seconds where PID = %d\n", i/10, jp->pid);
				break;
			}
			usleep(100000);
		}
		if(i == maxtry) {
			printf("Redir process in test is still a zombie after waiting %d seconds\n", i/10);
		}
		jp->pid = -1;
	}

	if ( jp->from_stdout >= 0 )
		close(jp->from_stdout);

	if ( jp->to_stdin >= 0 )
		close(jp->to_stdin);

	if ( jp->from_stderr >= 0 )
		close(jp->from_stderr);

	free(jp->program);
	free(jp);
}
