/**
 * @file
 * @brief This is a very simple squid 2.5 test case.  The requests are always allowed.
 *
 * Start a server that simply echo's the request in the reply
 * Open the redir
 * Send a request
 * Make sure we get the exact same thing back
 * Repeat a few times
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

static void connect_callback(struct client_socket* s)
{
	printf("%p: Connected\n", s);
}

static void recv_allow_callback(struct client_socket* s)
{
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);
		s->rxlen=0;
	}
}

static void recv_deny_callback(struct client_socket* s)
{
	char buffer[22];
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		strcpy(buffer, "http://denycallback/\n");
		socketserver_send(s, buffer, strlen(buffer));
		s->rxlen=0;
	}
}

static int recv_allowdeny_counter=0;
static void recv_allowdeny_callback(struct client_socket* s)
{
	char buffer[22];
	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		if ( recv_allowdeny_counter % 2 == 0 ) {
			// echo it back
			socketserver_send(s, s->rxbuffer, s->rxlen);
		}
		else {
			strcpy(buffer, "http://denycallback/\n");
			socketserver_send(s, buffer, strlen(buffer));
		}
		s->rxlen=0;
	}
}

static void disconnect_callback(struct client_socket* s)
{
	printf("%p: Disconnect\n", s);
}

/**
 * @brief This will start a server that will echo back what it got
 *
 * Uses the recv_allow_callback
 *
 */
static void simple_squid25_allow_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_allow_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, sizeof(buffer));
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, sendbuffer) ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void simple_squid25_deny_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_deny_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, sizeof(buffer));
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( strcmp(buffer, "http://denycallback/") ) {
			fprintf(stderr, "ret = %d buffer %s != http://denycallback/\n", ret, buffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void simple_squid25_allowdeny_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_allowdeny_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	recv_allowdeny_counter=0;
	for(i=0; i<100; i++) {
		recv_allowdeny_counter++;
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, sizeof(buffer));
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		if ( recv_allowdeny_counter % 2 == 0 ) {
			if ( strcmp(buffer, sendbuffer) ) {
				fprintf(stderr, "ret = %d buffer %s != %s\n", ret, buffer, sendbuffer);
				CU_FAIL("We did not get the response we expected.");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
		}
		else {
			if ( strcmp(buffer, "http://denycallback/") ) {
				fprintf(stderr, "ret = %d buffer %s != http://denycallback/\n", ret, buffer);
				CU_FAIL("We did not get the response we expected.");
				socketserver_signal_shutdown(svr);
				socketserver_waitfor_shutdown(svr);
				jclose(fp);
				return;
			}
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

static void recv_denywithurl_callback(struct client_socket* s)
{
	char buffer[1024];
	char* ptr;

	printf("%p: Data %s", s, s->rxbuffer);
	if ( strchr(s->rxbuffer, '\n' ) ) {
		strcpy(buffer, "http://deny/?url=");
		// remove after the first space
		ptr=strchr(s->rxbuffer, ' ' );
		if ( ptr ) {
			*ptr='\0';
		}
		strcat(buffer, s->rxbuffer);
		strcat(buffer, "\n");
		// echo it back
		socketserver_send(s, buffer, strlen(buffer));
		s->rxlen=0;
	}
}

static void simple_squid25_denywithurl_test()
{
	struct socketserver* svr;
	JFILE* fp;
	int i;
	char sendbuffer[256];
	char buffer[256];
	int ret;
	char* ptr;

	// Create a new amazing server
	svr = socketserver_create(3435, connect_callback, recv_denywithurl_callback, disconnect_callback);

	printf("%s(): Waiting for server startup\n", __FUNCTION__);

	fp = jopen("../redir localhost -f http://failcloseurl");
	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr);
		socketserver_waitfor_shutdown(svr);
		return;
	}

	for(i=0; i<100; i++) {
		sprintf(sendbuffer, "http://google.ca/%d -/- - -\n", i);
		ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));
		ret=read(fp->from_stdout, buffer, sizeof(buffer));
		if ( ret >= 0 ) {
			buffer[ret]='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(sendbuffer, ' ');
		*ptr='\0';

		// truncate sendbuffer at first space
		ptr=strchr(buffer, '\n');
		if ( ptr ) {
			*ptr='\0';
		}

		// truncate sendbuffer at first space
		ptr=strchr(buffer, ' ');
		if ( ptr ) {
			*ptr='\0';
		}

		// Must find the original url in the deny url
		// must find http://deny/?url= at the start
		if ( strstr(buffer, sendbuffer) == NULL || strncmp(buffer, "http://deny/?url=", 17) != 0 ) {
			fprintf(stderr, "ret = %d buffer %s != sendbuffer %s\n", ret, buffer, sendbuffer);
			CU_FAIL("We did not get the response we expected.");
			socketserver_signal_shutdown(svr);
			socketserver_waitfor_shutdown(svr);
			jclose(fp);
			return;
		}
	}
	

	socketserver_signal_shutdown(svr);
	socketserver_waitfor_shutdown(svr);
	jclose(fp);

	CU_PASS("The Simple Squid Redir test has passed");
	return;
}

int simple_squid25_test_init()
{
	CU_pSuite pSuite = NULL;

	// This is our test suite for test1
	pSuite = CU_add_suite("Simple Squid 2.5 Tests", NULL, NULL);
	if (NULL == pSuite) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Allow Test", simple_squid25_allow_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Add a test to this test suite
	if ( (NULL == CU_add_test(pSuite, "Deny Test", simple_squid25_deny_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Allow and deny test
	if ( (NULL == CU_add_test(pSuite, "Allow and Deny Test", simple_squid25_allowdeny_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	// Deny with the URL as well.. more amazing
	if ( (NULL == CU_add_test(pSuite, "Deny Test with original URL in Deny URL", simple_squid25_denywithurl_test)) ) {
		CU_cleanup_registry();
		return CU_get_error();
	}
	return 0;
}
