/**
 * @brief Check some simple examples of various squid requests
 *
 * Make sure allow and deny of all the different kinds of basic squid
 * requests are handled.
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "CUnit/Basic.h"
#include "jopen.h"
#include "socketserver.h"

static size_t simple_server_rcv_count;
// set this to make the server deny requests
// Set to NULL if server should echo first field
static const char *simple_server_deny;

/* @brief Info about the mock server.
 *
 * Static so init and cleanup can both get to it.
 */
static struct socketserver *svr1;

/* @brief Pointer to the redir process
 *
 * Static so init and cleanup can both get to it.
 */
static JFILE *fp;

/**
 * @brief the fake server calls this for every line received
 *
 * Receive data
 * Confirm that it looks uncorrupted
 * Increment a counter of lines of uncorrupted data received
 * Respond as indiciated by simple_server_deny global
 **/
static void simple_server(struct client_socket* s)
{
	int field = 0;
	char *rxptr;
	char *fieldstart;
	int resplen;

	// To be uncorrupted we need a line feed
	if (strchr (s->rxbuffer, '\n') == NULL)
	{
		// no line feed means just return and hope that
		// our framework does another read and gets us
		// some more data to chew on
		return;
	}

	// Now go through each field and count them
	// and make sure they're all non-zero
	rxptr = s->rxbuffer;
	fieldstart = rxptr;
	while (rxptr - s->rxbuffer < s->rxlen)
	{
		rxptr++;
		// found end of a field?
		if (*rxptr == ' ' || *rxptr == '\n')
		{
			// terminate the field
			*rxptr = '\0';
			// verify non-zero length
			if (*fieldstart == '\0')
			{
				s->rxlen = 0;
				return;
			}
			fieldstart = rxptr + 1;
			field++;
		}
	}

	// Did we find expected number of fields?
	if (field != 4)
	{
		s->rxlen=0;
		return;
	}

	// Increment success counter
	simple_server_rcv_count++;
	// Respond as desired by the tester.  Add a \n
	// ASSUME txbuffer is big enough
	if (simple_server_deny != NULL)
	{
		resplen = sprintf (s->txbuffer, "%s\n", simple_server_deny);
	}
	else
	{
		resplen = sprintf (s->txbuffer, "%s\n", s->rxbuffer);
	}
	socketserver_send(s, s->txbuffer, resplen);
	s->rxlen=0;
}

/**
 * @brief helper to send to stdin and get stdout
 *
 * Call this function to write to stdin and get a response from stdout
 *
 * @param jp an initialized redir JFILE
 * @param input a string to send to stdin
 * @param expected_out the expected output string.  NULL if should equal first field of input
 *
 * @returns 1 if expected response is received.
 **/
static int simple_io (JFILE *jp, const char *input, const char *expect_out)
{
	const size_t buff_len = 1024;
	char buff[buff_len];
	int len;

	// Write line to send to redir
	len = snprintf (buff, buff_len, "%s\n", input);
	if (len < 0 || len >= (int)buff_len)
	{
		fprintf (stderr, "Error writing to stdin buff.  input too big? (%s)\n", input);
		return 0;
	}
	// Send the line to stdin
	if (write (jp->to_stdin, buff, len) <= 0)
	{
		fprintf (stderr, "Error writing to stdin\n");
		return 0;
	}
	// Read response from stdout
	if ((len = read (jp->from_stdout, buff, buff_len)) <= 0)
	{
		fprintf (stderr, "Error reading stdout\n");
		return 0;
	}
	// Null terminate output string.  Get rid of \n
	CU_ASSERT_FATAL(buff[len - 1] == '\n');
	buff[len - 1] = '\0';

	// Compare for expected output
	if (expect_out == NULL)
	{
		if (strchr (input, ' ') == NULL)
		{
			if (strcmp (buff, input) != 0)
			{
				fprintf (stderr, "Output not equal to input.  Got \"%s\"\n", buff);
				return 0;
			}
		}
		else
		{
			int len = strchr (input, ' ') - input;
			if (memcmp (buff, input, len) != 0 ||
			    buff[len] != '\0')
			{
				fprintf (stderr, "Output not equal to first field of input.  Got \"%s\"\n", buff);
				return 0;
			}
		}

	}
	else if (strcmp (buff, expect_out) != 0)
	{
		fprintf (stderr, "Didnt get expected output.  Got \"%s\"\n", buff);
		return 0;
	}
	return 1;
}

/**
 * @brief Create the server and run redir
 *
 * This function sets up the environment for the tests by creating our mock
 * server and running the redir.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int simple_rewrite_initialize()
{
	// Create the server
	svr1 = socketserver_create(12345, NULL, simple_server, NULL);

	// Username rewriting MUST be enabled with the -u option
	fp = jopen("../redir localhost:12345 -f http://failcloseurl");

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return -1;
	}

	return 0;
}

/**
 * @brief Clean up the test environment
 *
 * This function cleans up the environment by shutting down the server and
 * stopping the redir process.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int simple_rewrite_cleanup()
{
	jclose(fp);

	socketserver_signal_shutdown(svr1);
	socketserver_waitfor_shutdown(svr1);

	return 0;
}


/**
 * @brief Simple http tests
 **/
static void simple_http_test ()
{
	simple_server_rcv_count = 0;
	simple_server_deny = NULL;
	CU_ASSERT(simple_io (fp, "http://google.com -/- - -", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
	CU_ASSERT(simple_io (fp, "http://google.com/path -/- - -", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 2);
	CU_ASSERT(simple_io (fp, "http://google.com/path/search?q=fdskafdaskfdskafdasfdsa&other=fdsafdsafdsafsdafsd -/- - -", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 3);

	simple_server_rcv_count = 0;
	simple_server_deny = "http://redirectme.com";
	CU_ASSERT(simple_io (fp, "http://google.com -/- - -", "http://redirectme.com"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
	CU_ASSERT(simple_io (fp, "http://google.com/path -/- - -", "http://redirectme.com"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 2);
	CU_ASSERT(simple_io (fp, "http://google.com/path/search?q=fdskafdaskfdskafdasfdsa&other=fdsafdsafdsafsdafsd -/- - -", "http://redirectme.com"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 3);

	simple_server_rcv_count = 0;
	simple_server_deny = "";
	CU_ASSERT(simple_io (fp, "http://google.com -/- - -", ""));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
	CU_ASSERT(simple_io (fp, "http://google.com/path -/- - -", ""));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 2);
	CU_ASSERT(simple_io (fp, "http://google.com/path/search?q=fdskafdaskfdskafdasfdsa&other=fdsafdsafdsafsdafsd -/- - -", ""));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 3);
}


/**
 * @brief Simple https tests
 **/
static void simple_https_test ()
{
	simple_server_rcv_count = 0;
	simple_server_deny = NULL;
	CU_ASSERT(simple_io (fp, "google.com:443 -/- - CONNECT", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
	CU_ASSERT(simple_io (fp, "google.com:9383 -/- - CONNECT", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 2);

	simple_server_rcv_count = 0;
	simple_server_deny = "denyhost:443";
	CU_ASSERT(simple_io (fp, "google.com:443 -/- - CONNECT", "denyhost:443"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
	CU_ASSERT(simple_io (fp, "google.com:9383 -/- - CONNECT", "denyhost:443"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 2);
}



/**
 * @brief Simple ftp tests
 **/
static void simple_ftp_test ()
{
	simple_server_rcv_count = 0;
	simple_server_deny = NULL;
	CU_ASSERT(simple_io (fp, "ftp://ftp.microsoft.com -/- - CONNECT", NULL));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);

	simple_server_rcv_count = 0;
	simple_server_deny = "denyftp:21";
	CU_ASSERT(simple_io (fp, "ftp://ftp.microsoft.com -/- - CONNECT", "denyftp:21"));
	CU_ASSERT_EQUAL(simple_server_rcv_count, 1);
}




int simple_test_init (void)
{
	CU_TestInfo simple_tests[] =
	   {
		{"simple http", simple_http_test},
		{"simple https", simple_https_test},
		{"simple ftp", simple_ftp_test},
		CU_TEST_INFO_NULL
	   };

	CU_SuiteInfo simple_suites[] =
	   {
		{ "simple rewrite test", simple_rewrite_initialize, simple_rewrite_cleanup, simple_tests },
		CU_SUITE_INFO_NULL
	   };

	if (CU_register_suites(simple_suites) != CUE_SUCCESS)
	{
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
