/**
 * @file
 * @brief Test client username character replacement and username domain mangling
 *
 * The redir does two separate types of client username mangling:
 *
 *     1. Replace characters that the NSD does not support.
 *     2. Change "domain\user" to "user@domain".
 *
 * We test both types of mangling in this suite.
 *
 *  These are the characters that the NSD supports in client usernames:
 *     # + @ _ . | \ ' = -
 *     A B C D E F G H I J K L M N O P Q R S T U V W X Y Z 
 *     a b c d e f g h i j k l m n o p q r s t u v w x y z 
 *     0 1 2 3 4 5 6 7 8 9
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "CUnit/Basic.h"

#include "jopen.h"
#include "socketserver.h"

// Arbitrarily-sized length for username we receive from redir
#define USERNAME_REWRITE_MAX_USERNAME_LEN 256

/**
 * Calling the QUOTE() macro will replace its argument (which must be
 * #define'd) with a quoted string of the value of that argument. Indirect
 * macro quoting expansion.
 *
 * They're used in the sscanf, where you have to specify the maximum field
 * width as part of the format string.
 */
#define _QUOTE(arg) #arg
#define QUOTE(arg) _QUOTE(arg)

/* @brief Info about the mock server.
 *
 * Static so init and cleanup can both get to it.
 */
static struct socketserver *svr1;

/* @brief Pointer to the redir process
 *
 * Static so init and cleanup can both get to it.
 */
static JFILE *fp;


/**
 * @brief Set up information about the socket connection
 *
 * This callback does nothing at present.
 */
static void connect_callback(__attribute__((unused)) struct client_socket* s)
{
	// No setup required
}

/**
 * @brief Allow the request in the socket rx buffer
 *
 * @param s the socket and associated buffers and statistics
 */
static void recv_allow_callback(struct client_socket* s)
{
	// Make sure the buffer has the whole request
	if ( strchr(s->rxbuffer, '\n' ) ) {
		// echo it back
		socketserver_send(s, s->rxbuffer, s->rxlen);

		// Consume the data
		s->rxlen = 0;
	}
}

/**
 * @brief Do cleanup before socket disconnect
 *
 * This callback does nothing at present.
 */
static void disconnect_callback(__attribute__((unused)) struct client_socket* s)
{
	// No cleanup required
}

/**
 * @brief Create the server and run redir
 *
 * This function sets up the environment for the tests by creating our mock
 * server and running the redir.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int username_rewrite_initialize()
{
	// Create the server and always allow the requests
	svr1 = socketserver_create(12345, connect_callback, recv_allow_callback, disconnect_callback);

	// Username rewriting MUST be enabled with the -u option
	fp = jopen("../redir localhost:12345 -f http://failcloseurl -u -G");

	if ( ! fp ) {
		fprintf(stderr, "Could not open ../redir\n");
		socketserver_signal_shutdown(svr1);
		socketserver_waitfor_shutdown(svr1);
		return -1;
	}

	return 0;
}

/**
 * @brief Clean up the test environment
 *
 * This function cleans up the environment by shutting down the server and
 * stopping the redir process.
 *
 * @return 0 if initialization was successful
 *        -1 on error
 */
static int username_rewrite_cleanup()
{
	jclose(fp);

	socketserver_signal_shutdown(svr1);
	socketserver_waitfor_shutdown(svr1);

	return 0;
}

/**
 * @brief Send a request to the server with a specific username
 *
 * @param input_username the username to send to the redir
 * @param expected_username the username we expect the redir to send to the
 * policy server
 *
 * @return 0 if we got the username we expected
 *        -1 if we failed (i.e. didn't get the username we expected)
 */
static int username_rewrite_send_request(const char *input_username, const char *expected_username)
{
	char sendbuffer[256];
	char buffer[256];
	char received_output[USERNAME_REWRITE_MAX_USERNAME_LEN];
	int ret;

	// Create the request that includes the input username
	sprintf(sendbuffer, "http://google.ca/ -/- %s -\n", input_username);

	// Send the request to the redir
	ret = write(fp->to_stdin, sendbuffer, strlen(sendbuffer));

	// Get the request from the redir to the policy server
	ret=read(fp->from_stdout, buffer, sizeof(buffer));
	if ( ret >= 0 ) {
		buffer[ret]='\0';
	}

	if ( ret < 0 ) {
		fprintf (stderr, "error reading from stdout\n");
		return -1;
	}


	// Get only the username from the request, and limit its size
	if ( sscanf(buffer, "%*s %*s %" QUOTE(USERNAME_REWRITE_MAX_USERNAME_LEN) "s %*s %*s", received_output) == 0 ) {
		// If we failed to convert in sscanf, terminate the username
		received_output[0] = '\0';
	}

	// Error if the output does not match what's expected
	if ( strcmp(received_output, expected_username) != 0) {
		//fprintf(stderr, "Failed: expected '%s', received '%s'\n", expected_username, received_output);
		return -1;
	}

	return 0;
}


// These tests verify that valid characters are not replaced

static void replace_chars_uppercase_test()
{
	CU_ASSERT(username_rewrite_send_request("ABCDEFGHIJKLM", "abcdefghijklm") == 0);
	CU_ASSERT(username_rewrite_send_request("NOPQRSTUVWXYZ", "nopqrstuvwxyz") == 0);
}

static void noreplace_chars_lowercase_test()
{
	CU_ASSERT(username_rewrite_send_request("abcdefghijklm", "abcdefghijklm") == 0);
	CU_ASSERT(username_rewrite_send_request("nopqrstuvwxyz", "nopqrstuvwxyz") == 0);
}

static void noreplace_chars_numbers_test()
{
	CU_ASSERT(username_rewrite_send_request("0123456789", "0123456789") == 0);
}

/**
 * @brief Number signs are valid and should NOT be replaced
 */
static void noreplace_number_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("#username", "#username") == 0);
	CU_ASSERT(username_rewrite_send_request("user#name", "user#name") == 0);
	CU_ASSERT(username_rewrite_send_request("username#", "username#") == 0);
}

/**
 * @brief Plus signs are valid and should NOT be replaced
 */
static void noreplace_plus_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("+username", "+username") == 0);
	CU_ASSERT(username_rewrite_send_request("user+name", "user+name") == 0);
	CU_ASSERT(username_rewrite_send_request("username+", "username+") == 0);
}

/**
 * @brief At signs are valid and should NOT be replaced
 */
static void noreplace_at_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("@username", "@username") == 0);
	CU_ASSERT(username_rewrite_send_request("user@name", "user@name") == 0);
	CU_ASSERT(username_rewrite_send_request("username@", "username@") == 0);
}

/**
 * @brief Underscores are valid and should NOT be replaced
 */
static void noreplace_underscore_test()
{
	CU_ASSERT(username_rewrite_send_request("_username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user_name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username_", "username_") == 0);
}

/**
 * @brief Periods are valid and should NOT be replaced
 */
static void noreplace_period_test()
{
	CU_ASSERT(username_rewrite_send_request(".username", ".username") == 0);
	CU_ASSERT(username_rewrite_send_request("user.name", "user.name") == 0);
	CU_ASSERT(username_rewrite_send_request("username.", "username.") == 0);
}

/**
 * @brief Pipes are valid and should NOT be replaced
 */
static void noreplace_pipe_test()
{
	CU_ASSERT(username_rewrite_send_request("|username", "|username") == 0);
	CU_ASSERT(username_rewrite_send_request("user|name", "user|name") == 0);
	CU_ASSERT(username_rewrite_send_request("username|", "username|") == 0);
}

/**
 * @brief Backslashes are valid and should NOT be replaced
 */
static void noreplace_backslash_test()
{
	CU_ASSERT(username_rewrite_send_request("\\\\username", "\\\\username") == 0);
	CU_ASSERT(username_rewrite_send_request("\\username", "\\username") == 0);
	CU_ASSERT(username_rewrite_send_request("username\\", "username\\") == 0);
	CU_ASSERT(username_rewrite_send_request("username\\\\", "username\\\\") == 0);
}

/**
 * @brief Single quotes are valid and should NOT be replaced
 */
static void noreplace_single_quote_test()
{
	CU_ASSERT(username_rewrite_send_request("'username", "'username") == 0);
	CU_ASSERT(username_rewrite_send_request("user'name", "user'name") == 0);
	CU_ASSERT(username_rewrite_send_request("username'", "username'") == 0);
}

/**
 * @brief Equals signs are valid and should NOT be replaced
 */
static void noreplace_equals_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("=username", "=username") == 0);
	CU_ASSERT(username_rewrite_send_request("user=name", "user=name") == 0);
	CU_ASSERT(username_rewrite_send_request("username=", "username=") == 0);
}

/**
 * @brief Hyphens are valid and should NOT be replaced
 */
static void noreplace_hyphen_test()
{
	CU_ASSERT(username_rewrite_send_request("-username", "-username") == 0);
	CU_ASSERT(username_rewrite_send_request("user-name", "user-name") == 0);
	CU_ASSERT(username_rewrite_send_request("username-", "username-") == 0);
}


// These tests verify that some common invalid characters are replaced with underscores

/**
 * @brief Test that exclamation marks are replaced with underscores
 */
static void replace_exclamation_mark_test()
{
	CU_ASSERT(username_rewrite_send_request("!username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user!name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username!", "username_") == 0);
}

/**
 * @brief Test that ampersands are replaced with underscores
 */
static void replace_ampersand_test()
{
	CU_ASSERT(username_rewrite_send_request("&username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user&name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username&", "username_") == 0);
}

/**
 * @brief Test that dollar signs are replaced with underscores
 */
static void replace_dollar_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("$username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user$name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username$", "username_") == 0);
}

/**
 * @brief Test that percent symbols are replaced with underscores
 */
static void replace_percent_symbol_test()
{
	CU_ASSERT(username_rewrite_send_request("\%username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user\%name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username\%", "username_") == 0);
}

/**
 * @brief Test that forward slashes are replaced with underscores
 */
static void replace_forward_slash_test()
{
	CU_ASSERT(username_rewrite_send_request("/username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user/name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username/", "username_") == 0);

}

/**
 * @brief Test that question marks are replaced with underscores
 */
static void replace_question_mark_test()
{
	CU_ASSERT(username_rewrite_send_request("?username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user?name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username?", "username_") == 0);

}

/**
 * @brief Test that less-than signs are replaced with underscores
 */
static void replace_lessthan_sign_test()
{
	CU_ASSERT(username_rewrite_send_request("<username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user<name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username<", "username_") == 0);

}

/**
 * @brief Test that greater-than signs are replaced with underscores
 */
static void replace_greaterthan_sign_test()
{
	CU_ASSERT(username_rewrite_send_request(">username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user>name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username>", "username_") == 0);

}

/**
 * @brief Test that carats are replaced with underscores
 */
static void replace_carat_test()
{
	CU_ASSERT(username_rewrite_send_request("^username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user^name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username^", "username_") == 0);

}

/**
 * @brief Test that asterisks are replaced with underscores
 */
static void replace_asterisk_test()
{
	CU_ASSERT(username_rewrite_send_request("*username", "_username") == 0);
	CU_ASSERT(username_rewrite_send_request("user*name", "user_name") == 0);
	CU_ASSERT(username_rewrite_send_request("username*", "username_") == 0);

}

// These tests verify that username domain mangling works

/**
 * @brief Test that a username in the form 'domain\user' is replaced with
 *        'user@domain'
 */
static void username_domain_mangle_test()
{
	CU_ASSERT(username_rewrite_send_request("domain\\user", "user@domain") == 0);
	CU_ASSERT(username_rewrite_send_request("domain%5cuser", "user@domain") == 0);
}

/**
 * @brief Test that we don't mangle something that almost looks like username
 *        domain but isn't
 */
static void username_domain_nomangle_test()
{
	CU_ASSERT(username_rewrite_send_request("domain\\\\user", "domain\\\\user") == 0);
}

/**
 * @brief Test that we mangle a username and domain that are both really long
 *
 *        There isn't a huge need for this test, granted, but it can't hurt.
 */
static void username_domain_mangle_long_test()
{
	CU_ASSERT(username_rewrite_send_request("longlonglonglonglonglonglonglongdomain\\longlonglonglonglonglonglonglonguser", "longlonglonglonglonglonglonglonguser@longlonglonglonglonglonglonglongdomain") == 0);
}

/**
 * @brief Register our test cases and test suite with CUnit
 *
 * @return 0 if we registered everything successfully
 *         nonzero otherwise (returns CU_ErrorCode errors)
 */
int username_rewrite_test_init()
{
	// Username rewrite tests
	CU_TestInfo username_rewrite_tests[] = {
		// Don't replace valid characters tests
		{ "Replace uppercase letters",      replace_chars_uppercase_test }, 
		{ "Don't replace lowercase letters",      noreplace_chars_lowercase_test }, 
		{ "Don't replace numbers",                noreplace_chars_numbers_test }, 
		{ "Don't replace number signs",           noreplace_number_sign_test }, 
		{ "Don't replace plus signs",             noreplace_plus_sign_test }, 
		{ "Don't replace at signs",               noreplace_at_sign_test }, 
		{ "Don't replace underscores",            noreplace_underscore_test }, 
		{ "Don't replace periods",                noreplace_period_test }, 
		{ "Don't replace pipes",                  noreplace_pipe_test }, 
		{ "Don't replace backslashes",            noreplace_backslash_test }, 
		{ "Don't replace single quotes",          noreplace_single_quote_test }, 
		{ "Don't replace equals signs",           noreplace_equals_sign_test }, 
		{ "Don't replace hyphens",                noreplace_hyphen_test }, 
		
		// Replace invalid characters tests
		{ "Replace exclamation mark",             replace_exclamation_mark_test },
		{ "Replace ampersand",                    replace_ampersand_test },
		{ "Replace dollar sign",                  replace_dollar_sign_test },
		{ "Replace percent symbol",               replace_percent_symbol_test },
		{ "Replace forward slash",                replace_forward_slash_test },
		{ "Replace question mark",                replace_question_mark_test },
		{ "Replace less-than sign",               replace_lessthan_sign_test },
		{ "Replace greater-than sign",            replace_greaterthan_sign_test },
		{ "Replace carat",                        replace_carat_test },
		{ "Replace asterisk",                     replace_asterisk_test },

		// Username domain mangling tests
		{ "Mangle username domain",               username_domain_mangle_test },
		{ "Don't mangle invalid username domain", username_domain_nomangle_test },
		{ "Mangle long username and domain",      username_domain_mangle_long_test },

		CU_TEST_INFO_NULL
	};

	// List our test suites, which contain tests
	CU_SuiteInfo username_rewrite_suites[] = {
		{ "Username Rewrite Test Suite", username_rewrite_initialize, username_rewrite_cleanup, username_rewrite_tests },
		CU_SUITE_INFO_NULL
	};
	
	// Register our test suites with CUnit
	if (CU_register_suites(username_rewrite_suites) != CUE_SUCCESS) {
		CU_cleanup_registry();
		return CU_get_error();
	}

	return 0;
}
