/**
 * @file
 * @brief Utils Header
 *
 *
 * Netsweeper Squid URL Redirector
 * This application is designed to work with the Netsweeper Internet Content Filtering System.
 *
 * Copyright 2010 Netsweeper Inc. All Rights Reserved.
 * Netsweeper, Netsweeper Inc. is either a copyrighted or registered trademark of Netsweeper Inc.
 */
#ifndef _UTILS_H
#define _UTILS_H

/** @brief A data structure to keep a locked, thread safe count */
typedef struct _global_count {
	/** @brief The current count we are keeping track of */
	int num;

	/** @brief The lock used to increment the count */
	pthread_mutex_t lock;
} GlobalCount;

/**
 * @brief GlobalCount initialization macro
 *
 * You can use this in the following method
 *	GlobalCount test = GLOBAL_COUNT_INIT;
 */
#define GLOBAL_COUNT_INIT {0,PTHREAD_MUTEX_INITIALIZER}

/**
 * @brief Data structure used for our queue
 *
 * You should use the following functions to access this data structure,
 *
 *	ReqQueue* redir_init_queue();
 *	void redir_queue_push(ReqQueue* queue, void* req);
 *	Request* redir_queue_pop(ReqQueue* queue);
 *	Request* redir_queue_walk(ReqQueue* queue);
 *	int redir_queue_size(ReqQueue* queue);
 */
typedef struct _req_queue {
	/** @brief The first item in the queue */
	Request*	head;
	/** @brief The last item in the queue, can be also the first */
	Request*	tail;
	/** @brief The number of items in the queue */
	int		size;
	/** @brief The lock to push/pop items */
	pthread_mutex_t	lock;
	/** @brief Signal used to wait for items to come into the queue */
	pthread_cond_t	cond;
} ReqQueue;

// Global counters
void redir_st_up();
void redir_st_down();
int redir_get_servers();

void redir_st_run();
void redir_st_exit();
int redir_get_threads();

// Queue API
ReqQueue* redir_init_queue();
void redir_free_queue(ReqQueue* queue);
void redir_queue_push(ReqQueue* queue, void* req);
Request* redir_queue_pop(ReqQueue* queue, Connection *conn, int timeout);
Request* redir_queue_walk(ReqQueue* queue);
int redir_queue_size(ReqQueue* queue);

// Request Dispatcher API
int redir_dispatcher_init(GlobalConf *conf);
void redir_dispatcher_terminate();
void redir_dispatch_request(Request *request, int tohead);
Request* redir_pop_request(Connection *conn);
void redir_start_flush_request_queue();
void redir_do_flush_request_queue(int force);
void redir_check_dispatch_space();

void redir_dispatcher_broadcast();

void * queue_flush_thread(void* arg);

#endif
